# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

'''
Indian GPA example from:

Discrete-Continuous Mixtures in Probabilistic Programming: Generalized
Semantics and Inference Algorithms, Wu et. al., ICML 2018.
https://arxiv.org/pdf/1806.02027.pdf
'''

import pytest

from sppl.compilers.ast_to_spn import Id
from sppl.compilers.ast_to_spn import IfElse
from sppl.compilers.ast_to_spn import Sample
from sppl.compilers.ast_to_spn import Sequence
from sppl.compilers.sppl_to_python import SPPL_Compiler
from sppl.distributions import atomic
from sppl.distributions import choice
from sppl.distributions import uniform
from sppl.math_util import allclose
from sppl.sets import Interval
from sppl.spn import ExposedSumSPN

Nationality = Id('Nationality')
Perfect     = Id('Perfect')
GPA         = Id('GPA')

def model_no_latents():
    return \
        0.5 * ( # American student
            0.99 * (GPA >> uniform(loc=0, scale=4)) | \
            0.01 * (GPA >> atomic(loc=4))) | \
        0.5 * ( # Indian student
            0.99 * (GPA >> uniform(loc=0, scale=10)) | \
            0.01 * (GPA >> atomic(loc=10)))

def model_exposed():
    return ExposedSumSPN(
        spn_weights=(Nationality >> choice({'India': 0.5, 'USA': 0.5})),
        children={
            # American student.
            'USA': ExposedSumSPN(
                spn_weights=(Perfect >> choice({'True': 0.01, 'False': 0.99})),
                children={
                    'False'   : GPA >> uniform(loc=0, scale=4),
                    'True'    : GPA >> atomic(loc=4),
                }),
            # Indian student.
            'India': ExposedSumSPN(
                spn_weights=(Perfect >> choice({'True': 0.01, 'False': 0.99})),
                children={
                    'False'   : GPA >> uniform(loc=0, scale=10),
                    'True'    : GPA >> atomic(loc=10),
                })},
        )

def model_ifelse_exhuastive():
    command = Sequence(
        Sample(Nationality, choice({'India': 0.5, 'USA': 0.5})),
        Sample(Perfect,     choice({'True': 0.01, 'False': 0.99})),
        IfElse(
            (Nationality << {'India'}) & (Perfect << {'False'}),
                Sample(GPA, uniform(loc=0, scale=10))
            ,
            (Nationality << {'India'}) & (Perfect << {'True'}),
                Sample(GPA, atomic(loc=10))
            ,
            (Nationality << {'USA'}) & (Perfect << {'False'}),
                Sample(GPA, uniform(loc=0, scale=4))
            ,
            (Nationality << {'USA'}) & (Perfect << {'True'}),
                Sample(GPA, atomic(loc=4))))
    return command.interpret()

def model_ifelse_non_exhuastive():
    Nationality = Id('Nationality')
    Perfect     = Id('Perfect')
    GPA         = Id('GPA')
    command = Sequence(
        Sample(Nationality, choice({'India': 0.5, 'USA': 0.5})),
        Sample(Perfect,     choice({'True': 0.01, 'False': 0.99})),
        IfElse(
            (Nationality << {'India'}) & (Perfect << {'False'}),
                Sample(GPA, uniform(loc=0, scale=10))
            ,
            (Nationality << {'India'}) & (Perfect << {'True'}),
                Sample(GPA, atomic(loc=10))
            ,
            (Nationality << {'USA'}) & (Perfect << {'False'}),
                Sample(GPA, uniform(loc=0, scale=4))
            ,
            True,
                Sample(GPA, atomic(loc=4))))
    return command.interpret()

def model_ifelse_nested():
    Nationality = Id('Nationality')
    Perfect     = Id('Perfect')
    GPA         = Id('GPA')
    command = Sequence(
        Sample(Nationality, choice({'India': 0.5, 'USA': 0.5})),
        Sample(Perfect,     choice({'True': 0.01, 'False': 0.99})),
        IfElse(
            Nationality << {'India'},
                IfElse(
                    Perfect << {'True'},    Sample(GPA, atomic(loc=10)),
                    Perfect << {'False'},   Sample(GPA, uniform(scale=10)),
                ),
            Nationality << {'USA'},
                IfElse(
                    Perfect << {'True'},    Sample(GPA, atomic(loc=4)),
                    Perfect << {'False'},   Sample(GPA, uniform(scale=4)),
                )))
    return command.interpret()

def model_perfect_nested():
    Nationality = Id('Nationality')
    Perfect     = Id('Perfect')
    GPA         = Id('GPA')
    command = Sequence(
        Sample(Nationality, choice({'India': 0.5, 'USA': 0.5})),
        IfElse(
            Nationality << {'India'}, Sequence(
                    Sample(Perfect, choice({'True': 0.01, 'False': 0.99})),
                    IfElse(
                        Perfect << {'True'},    Sample(GPA, atomic(loc=10)),
                        True,                   Sample(GPA, uniform(scale=10))
                    )),
            Nationality << {'USA'}, Sequence(
                Sample(Perfect, choice({'True': 0.01, 'False': 0.99})),
                IfElse(
                    Perfect << {'True'},    Sample(GPA, atomic(loc=4)),
                    True,                   Sample(GPA, uniform(scale=4)),
                ))))
    return command.interpret()

def model_ifelse_exhuastive_compiled():
    compiler = SPPL_Compiler('''
Nationality   ~= choice({'India': 0.5, 'USA': 0.5})
Perfect       ~= choice({'True': 0.01, 'False': 0.99})
if (Nationality == 'India') & (Perfect == 'False'):
    GPA ~= uniform(loc=0, scale=10)
elif (Nationality == 'India') & (Perfect == 'True'):
    GPA ~= atomic(loc=10)
elif (Nationality == 'USA') & (Perfect == 'False'):
    GPA ~= uniform(loc=0, scale=4)
elif (Nationality == 'USA') & (Perfect == 'True'):
    GPA ~= atomic(loc=4)
    ''')
    namespace = compiler.execute_module()
    return namespace.model

def model_ifelse_non_exhuastive_compiled():
    compiler = SPPL_Compiler('''
Nationality   ~= choice({'India': 0.5, 'USA': 0.5})
Perfect       ~= choice({'True': 0.01, 'False': 0.99})
if (Nationality == 'India') & (Perfect == 'False'):
    GPA ~= uniform(loc=0, scale=10)
elif (Nationality == 'India') & (Perfect == 'True'):
    GPA ~= atomic(loc=10)
elif (Nationality == 'USA') & (Perfect == 'False'):
    GPA ~= uniform(loc=0, scale=4)
else:
    GPA ~= atomic(loc=4)
    ''')
    namespace = compiler.execute_module()
    return namespace.model

def model_ifelse_nested_compiled():
    compiler = SPPL_Compiler('''
Nationality   ~= choice({'India': 0.5, 'USA': 0.5})
Perfect       ~= choice({'True': 0.01, 'False': 0.99})
if (Nationality == 'India'):
    if (Perfect == 'False'):
        GPA ~= uniform(loc=0, scale=10)
    else:
        GPA ~= atomic(loc=10)
elif (Nationality == 'USA'):
    if (Perfect == 'False'):
        GPA ~= uniform(loc=0, scale=4)
    elif (Perfect == 'True'):
        GPA ~= atomic(loc=4)
    ''')
    namespace = compiler.execute_module()
    return namespace.model

def model_perfect_nested_compiled():
    compiler = SPPL_Compiler('''
Nationality   ~= choice({'India': 0.5, 'USA': 0.5})
if (Nationality == 'India'):
    Perfect       ~= choice({'True': 0.01, 'False': 0.99})
    if (Perfect == 'False'):
        GPA ~= uniform(loc=0, scale=10)
    else:
        GPA ~= atomic(loc=10)
elif (Nationality == 'USA'):
    Perfect       ~= choice({'True': 0.01, 'False': 0.99})
    if (Perfect == 'False'):
        GPA ~= uniform(loc=0, scale=4)
    else:
        GPA ~= atomic(loc=4)
    ''')
    namespace = compiler.execute_module()
    return namespace.model

@pytest.mark.parametrize('get_model', [
    # Manual
    model_no_latents,
    model_exposed,
    # Interpreter
    model_ifelse_exhuastive,
    model_ifelse_non_exhuastive,
    model_ifelse_nested,
    model_perfect_nested,
    # Compiler
    model_ifelse_exhuastive_compiled,
    model_ifelse_non_exhuastive_compiled,
    model_ifelse_nested_compiled,
    model_perfect_nested_compiled,
])
def test_prior(get_model):
    model = get_model()
    GPA = Id('GPA')
    assert allclose(model.prob(GPA << {10}), 0.5*0.01)
    assert allclose(model.prob(GPA << {4}), 0.5*0.01)
    assert allclose(model.prob(GPA << {5}), 0)
    assert allclose(model.prob(GPA << {1}), 0)

    assert allclose(model.prob((2 < GPA) < 4),
        0.5*0.99*0.5 + 0.5*0.99*0.2)
    assert allclose(model.prob((2 <= GPA) < 4),
        0.5*0.99*0.5 + 0.5*0.99*0.2)
    assert allclose(model.prob((2 < GPA) <= 4),
        0.5*(0.99*0.5 + 0.01) + 0.5*0.99*0.2)
    assert allclose(model.prob((2 < GPA) <= 8),
        0.5*(0.99*0.5 + 0.01) + 0.5*0.99*0.6)
    assert allclose(model.prob((2 < GPA) < 10),
        0.5*(0.99*0.5 + 0.01) + 0.5*0.99*0.8)
    assert allclose(model.prob((2 < GPA) <= 10),
        0.5*(0.99*0.5 + 0.01) + 0.5*(0.99*0.8 + 0.01))

    assert allclose(model.prob(((2 <= GPA) < 4) | (7 < GPA)),
        (0.5*0.99*0.5 + 0.5*0.99*0.2) + (0.5*(0.99*0.3 + 0.01)))

    assert allclose(model.prob(((2 <= GPA) < 4) & (7 < GPA)), 0)

def test_condition():
    model = model_no_latents()
    GPA = Id('GPA')
    model_condition = model.condition(GPA << {4} | GPA << {10})
    assert len(model_condition.children) == 2
    assert model_condition.children[0].support == Interval.Ropen(4, 5)
    assert model_condition.children[1].support == Interval.Ropen(10, 11)

    model_condition = model.condition((0 < GPA < 4))
    assert len(model_condition.children) == 2
    assert model_condition.children[0].support \
        == model_condition.children[1].support
    assert allclose(
        model_condition.children[0].logprob(GPA < 1),
        model_condition.children[1].logprob(GPA < 1))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import scipy.stats
import sympy

class Distribution():
    def __rmul__(self, x):
        from .sym_util import sympify_number
        try:
            x_val = sympify_number(x)
            if not 0 < x_val < 1:
                raise ValueError('invalid weight %s' % (str(x),))
            return DistributionMix([self], [x])
        except TypeError:
            return NotImplemented

class NominalDistribution(Distribution):
    def __init__(self, dist):
        self.dist = dict(dist)
    def __call__(self, symbol):
        from .spn import NominalLeaf
        return NominalLeaf(symbol, self.dist)

choice = NominalDistribution

def floatify(x):
    try              : return float(x)
    except TypeError : return x

class RealDistribution(Distribution):
    # pylint: disable=not-callable
    # pylint: disable=multiple-statements
    dist = None
    constructor = None
    def __init__(self, *args, **kwargs):
        assert not args, 'Only keyword arguments allowed for %s' % (self.dist.name,)
        self.kwargs = {k: floatify(v) for k, v in kwargs.items()}
    def __call__(self, symbol):
        domain = self.get_domain()
        return self.constructor(symbol, self.dist(**self.kwargs), domain)
    def get_domain(self):
        raise NotImplementedError()

class DistributionMix():
    """Weighted mixture of SPNs that do not yet sum to unity."""
    def __init__(self, distributions, weights):
        self.distributions = distributions
        self.weights = weights
    def __call__(self, symbol):
        from math import log
        from .spn import SumSPN
        distributions = [d(symbol) for d in self.distributions]
        weights = [log(w) for w in self.weights]
        return SumSPN(distributions, weights)

    def __or__(self, x):
        if not isinstance(x, DistributionMix):
            return NotImplemented
        weights = self.weights + x.weights
        cumsum = float(sum(weights))
        assert 0 < cumsum <= 1
        distributions = self.distributions + x.distributions
        return DistributionMix(distributions, weights)

# ==============================================================================
# ContinuousReal

from .sets import Interval
from .sets import Reals
from .sets import RealsNeg
from .sets import RealsPos
from .sets import inf as oo
from .spn import ContinuousLeaf

def RealsPosLoc(kwargs):
    if 'loc' in kwargs:
        return Interval(kwargs['loc'], oo)
    return RealsPos

def UnitIntervalLocScale(kwargs):
    loc = kwargs.get('loc', 0)
    scale = kwargs.get('scale', 1)
    return Interval(loc, loc + scale)

class ContinuousReal(RealDistribution):
    constructor = ContinuousLeaf

class alpha(ContinuousReal):
    """An alpha continuous random variable."""
    dist = scipy.stats.alpha
    def get_domain(self): return RealsPosLoc(self.kwargs)

class anglit(ContinuousReal):
    """An anglit continuous random variable."""
    dist = scipy.stats.anglit
    def get_domain(self): return Interval(-sympy.pi/4, sympy.pi/4)

class arcsine(ContinuousReal):
    """An arcsine continuous random variable."""
    dist = scipy.stats.arcsine
    def get_domain(self): return UnitIntervalLocScale(self.kwargs)

class argus(ContinuousReal):
    """Argus distribution"""
    dist = scipy.stats.argus
    def get_domain(self): return UnitIntervalLocScale(self.kwargs)

class beta(ContinuousReal):
    """A beta continuous random variable."""
    dist = scipy.stats.beta
    def get_domain(self): return UnitIntervalLocScale(self.kwargs)

class betaprime(ContinuousReal):
    """A beta prime continuous random variable."""
    dist = scipy.stats.betaprime
    def get_domain(self): return RealsPosLoc(self.kwargs)

class bradford(ContinuousReal):
    """A Bradford continuous random variable."""
    dist = scipy.stats.bradford
    def get_domain(self): return UnitIntervalLocScale(self.kwargs)

class burr(ContinuousReal):
    """A Burr (Type III) continuous random variable."""
    dist = scipy.stats.burr
    def get_domain(self): return RealsPosLoc(self.kwargs)

class burr12(ContinuousReal):
    """A Burr (Type XII) continuous random variable."""
    dist = scipy.stats.burr12
    def get_domain(self): return RealsPosLoc(self.kwargs)

class cauchy(ContinuousReal):
    """A Cauchy continuous random variable."""
    dist = scipy.stats.cauchy
    def get_domain(self): return Reals

class chi(ContinuousReal):
    """A chi continuous random variable."""
    dist = scipy.stats.chi
    def get_domain(self): return RealsPosLoc(self.kwargs)

class chi2(ContinuousReal):
    """A chi-squared continuous random variable."""
    dist = scipy.stats.chi2
    def get_domain(self): return RealsPosLoc(self.kwargs)

class cosine(ContinuousReal):
    """A cosine continuous random variable."""
    dist = scipy.stats.cosine
    def get_domain(self): return Interval(-sympy.pi/2, sympy.pi/2)

class crystalball(ContinuousReal):
    """Crystalball distribution."""
    dist = scipy.stats.crystalball
    def get_domain(self): return Reals

class dgamma(ContinuousReal):
    """A double gamma continuous random variable."""
    dist = scipy.stats.dgamma
    def get_domain(self): return Reals

class dweibull(ContinuousReal):
    """A double Weibull continuous random variable."""
    dist = scipy.stats.dweibull
    def get_domain(self): return Reals

class erlang(ContinuousReal):
    """An Erlang continuous random variable."""
    dist = scipy.stats.erlang
    def get_domain(self): return RealsPosLoc(self.kwargs)

class expon(ContinuousReal):
    """An exponential continuous random variable."""
    dist = scipy.stats.expon
    def get_domain(self): return RealsPosLoc(self.kwargs)

class exponnorm(ContinuousReal):
    """An exponentially modified normal continuous random variable."""
    dist = scipy.stats.exponnorm
    def get_domain(self): return Reals

class exponweib(ContinuousReal):
    """An exponentiated Weibull continuous random variable."""
    dist = scipy.stats.exponweib
    def get_domain(self): return RealsPosLoc(self.kwargs)

class exponpow(ContinuousReal):
    """An exponential power continuous random variable."""
    dist = scipy.stats.exponpow
    def get_domain(self): return RealsPosLoc(self.kwargs)

class f(ContinuousReal):
    """An F continuous random variable."""
    dist = scipy.stats.f
    def get_domain(self): return RealsPosLoc(self.kwargs)

class fatiguelife(ContinuousReal):
    """A fatigue-life (Birnbaum-Saunders) continuous random variable."""
    dist = scipy.stats.fatiguelife
    def get_domain(self): return RealsPosLoc(self.kwargs)

class fisk(ContinuousReal):
    """A Fisk continuous random variable."""
    dist = scipy.stats.fisk
    def get_domain(self): return RealsPosLoc(self.kwargs)

class foldcauchy(ContinuousReal):
    """A folded Cauchy continuous random variable."""
    dist = scipy.stats.foldcauchy
    def get_domain(self): return RealsPosLoc(self.kwargs)

class foldnorm(ContinuousReal):
    """A folded normal continuous random variable."""
    dist = scipy.stats.foldnorm
    def get_domain(self): return RealsPosLoc(self.kwargs)

class frechet_r(ContinuousReal):
    """A Frechet right (or Weibull minimum) continuous random variable."""
    dist = scipy.stats.frechet_r
    def get_domain(self): return RealsPosLoc(self.kwargs)

class frechet_l(ContinuousReal):
    """A Frechet left (or Weibull maximum) continuous random variable."""
    dist = scipy.stats.frechet_l
    def get_domain(self): return RealsNeg

class genlogistic(ContinuousReal):
    """A generalized logistic continuous random variable."""
    dist = scipy.stats.genlogistic
    def get_domain(self): return RealsPosLoc(self.kwargs)

class gennorm(ContinuousReal):
    """A generalized normal continuous random variable."""
    dist = scipy.stats.gennorm
    def get_domain(self): return Reals

class genpareto(ContinuousReal):
    """A generalized Pareto continuous random variable."""
    dist = scipy.stats.genpareto
    def get_domain(self): return RealsPosLoc(self.kwargs)

class genexpon(ContinuousReal):
    """A generalized exponential continuous random variable."""
    dist = scipy.stats.genexpon
    def get_domain(self): return RealsPosLoc(self.kwargs)

class genextreme(ContinuousReal):
    """A generalized extreme value continuous random variable."""
    dist = scipy.stats.genextreme
    def get_domain(self):
        c = self.kwargs['c']
        if c == 0:
            return Reals
        elif c > 0:
            return Interval(-oo, 1/c)
        elif c < 0:
            return Interval(1/c, oo)
        assert False, 'Bad argument "c" for genextreme: %s' % (self.kwargs,)

class gausshyper(ContinuousReal):
    """A Gauss hypergeometric continuous random variable."""
    dist = scipy.stats.gausshyper
    def get_domain(self): return UnitIntervalLocScale(self.kwargs)

class gamma(ContinuousReal):
    """A gamma continuous random variable."""
    dist = scipy.stats.gamma
    def get_domain(self): return RealsPosLoc(self.kwargs)

class gengamma(ContinuousReal):
    """A generalized gamma continuous random variable."""
    dist = scipy.stats.gengamma
    def get_domain(self): return RealsPosLoc(self.kwargs)

class genhalflogistic(ContinuousReal):
    """A generalized half-logistic continuous random variable."""
    dist = scipy.stats.genhalflogistic
    def get_domain(self):
        assert self.kwargs['c'] > 0
        return Interval(0, 1./self.kwargs['c'])

class geninvgauss(ContinuousReal):
    """A Generalized Inverse Gaussian continuous random variable."""
    dist = scipy.stats.geninvgauss
    def get_domain(self): return RealsPosLoc(self.kwargs)

class gilbrat(ContinuousReal):
    """A Gilbrat continuous random variable."""
    dist = scipy.stats.gilbrat
    def get_domain(self): return RealsPosLoc(self.kwargs)

class gompertz(ContinuousReal):
    """A Gompertz (or truncated Gumbel) continuous random variable."""
    dist = scipy.stats.gompertz
    def get_domain(self): return RealsPosLoc(self.kwargs)

class gumbel_r(ContinuousReal):
    """A right-skewed Gumbel continuous random variable."""
    dist = scipy.stats.gumbel_r
    def get_domain(self): return Reals

class gumbel_l(ContinuousReal):
    """A left-skewed Gumbel continuous random variable."""
    dist = scipy.stats.gumbel_l
    def get_domain(self): return RealsPosLoc(self.kwargs)

class halfcauchy(ContinuousReal):
    """A Half-Cauchy continuous random variable."""
    dist = scipy.stats.halfcauchy
    def get_domain(self): return RealsPosLoc(self.kwargs)

class halflogistic(ContinuousReal):
    """A half-logistic continuous random variable."""
    dist = scipy.stats.halflogistic
    def get_domain(self): return RealsPosLoc(self.kwargs)

class halfnorm(ContinuousReal):
    """A half-normal continuous random variable."""
    dist = scipy.stats.halfnorm
    def get_domain(self): return RealsPosLoc(self.kwargs)

class halfgennorm(ContinuousReal):
    """The upper half of a generalized normal continuous random variable."""
    dist = scipy.stats.halfgennorm
    def get_domain(self): return RealsPosLoc(self.kwargs)

class hypsecant(ContinuousReal):
    """A hyperbolic secant continuous random variable."""
    dist = scipy.stats.hypsecant
    def get_domain(self): return Reals

class invgamma(ContinuousReal):
    """An inverted gamma continuous random variable."""
    dist = scipy.stats.invgamma
    def get_domain(self): return RealsPosLoc(self.kwargs)

class invgauss(ContinuousReal):
    """An inverse Gaussian continuous random variable."""
    dist = scipy.stats.invgauss
    def get_domain(self): return RealsPosLoc(self.kwargs)

class invweibull(ContinuousReal):
    """An inverted Weibull continuous random variable."""
    dist = scipy.stats.invweibull
    def get_domain(self): return RealsPosLoc(self.kwargs)

class johnsonsb(ContinuousReal):
    """A Johnson SB continuous random variable."""
    dist = scipy.stats.johnsonsb
    def get_domain(self): return UnitIntervalLocScale(self.kwargs)

class johnsonsu(ContinuousReal):
    """A Johnson SU continuous random variable."""
    dist = scipy.stats.johnsonsu
    def get_domain(self): return Reals

class kappa4(ContinuousReal):
    """Kappa 4 parameter distribution."""
    dist = scipy.stats.kappa4
    def get_domain(self): return Reals

class kappa3(ContinuousReal):
    """Kappa 3 parameter distribution."""
    dist = scipy.stats.kappa3
    def get_domain(self): return RealsPosLoc(self.kwargs)

class ksone(ContinuousReal):
    """General Kolmogorov-Smirnov one-sided test."""
    dist = scipy.stats.ksone
    def get_domain(self): return UnitIntervalLocScale(self.kwargs)

class kstwobign(ContinuousReal):
    """Kolmogorov-Smirnov two-sided test for large N."""
    dist = scipy.stats.kstwobign
    def get_domain(self): return Interval(0, sympy.sqrt(self.kwargs['n']))

class laplace(ContinuousReal):
    """A Laplace continuous random variable."""
    dist = scipy.stats.laplace
    def get_domain(self): return Reals

class levy(ContinuousReal):
    """A Levy continuous random variable."""
    dist = scipy.stats.levy
    def get_domain(self): return RealsPosLoc(self.kwargs)

class levy_l(ContinuousReal):
    """A left-skewed Levy continuous random variable."""
    dist = scipy.stats.levy_l
    def get_domain(self): return RealsNeg

class levy_stable(ContinuousReal):
    """A Levy-stable continuous random variable."""
    dist = scipy.stats.levy_stable
    def get_domain(self): return Reals

class logistic(ContinuousReal):
    """A logistic (or Sech-squared) continuous random variable."""
    dist = scipy.stats.logistic
    def get_domain(self): return Reals

class loggamma(ContinuousReal):
    """A log gamma continuous random variable."""
    dist = scipy.stats.loggamma
    def get_domain(self): return RealsNeg

class loglaplace(ContinuousReal):
    """A log-Laplace continuous random variable."""
    dist = scipy.stats.loglaplace
    def get_domain(self): return RealsPosLoc(self.kwargs)

class lognorm(ContinuousReal):
    """A lognormal continuous random variable."""
    dist = scipy.stats.lognorm
    def get_domain(self): return RealsPosLoc(self.kwargs)

class loguniform(ContinuousReal):
    """A loguniform or reciprocal continuous random variable."""
    dist = scipy.stats.loguniform
    def get_domain(self): return Interval(self.kwargs['a'], self.kwargs['b'])

class lomax(ContinuousReal):
    """A Lomax (Pareto of the second kind) continuous random variable."""
    dist = scipy.stats.lomax
    def get_domain(self): return RealsPosLoc(self.kwargs)

class maxwell(ContinuousReal):
    """A Maxwell continuous random variable."""
    dist = scipy.stats.maxwell
    def get_domain(self): return RealsPosLoc(self.kwargs)

class mielke(ContinuousReal):
    """A Mielke Beta-Kappa / Dagum continuous random variable."""
    dist = scipy.stats.mielke
    def get_domain(self): return RealsPosLoc(self.kwargs)

class moyal(ContinuousReal):
    """A Moyal continuous random variable."""
    dist = scipy.stats.moyal
    def get_domain(self): return Reals

class nakagami(ContinuousReal):
    """A Nakagami continuous random variable."""
    dist = scipy.stats.nakagami
    def get_domain(self): return RealsPosLoc(self.kwargs)

class ncx2(ContinuousReal):
    """A non-central chi-squared continuous random variable."""
    dist = scipy.stats.ncx2
    def get_domain(self): return RealsPosLoc(self.kwargs)

class ncf(ContinuousReal):
    """A non-central F distribution continuous random variable."""
    dist = scipy.stats.ncf
    def get_domain(self): return RealsPosLoc(self.kwargs)

class nct(ContinuousReal):
    """A non-central Student’s t continuous random variable."""
    dist = scipy.stats.nct
    def get_domain(self): return Reals

class norm(ContinuousReal):
    """A normal continuous random variable."""
    dist = scipy.stats.norm
    def get_domain(self): return Reals
normal = norm

class norminvgauss(ContinuousReal):
    """A normal Inverse Gaussian continuous random variable."""
    dist = scipy.stats.norminvgauss
    def get_domain(self): return Reals

class pareto(ContinuousReal):
    """A Pareto continuous random variable."""
    dist = scipy.stats.pareto
    def get_domain(self): return Interval(1, oo)

class pearson3(ContinuousReal):
    """A pearson type III continuous random variable."""
    dist = scipy.stats.pearson3
    def get_domain(self): return Reals

class powerlaw(ContinuousReal):
    """A power-function continuous random variable."""
    dist = scipy.stats.powerlaw
    def get_domain(self): return UnitIntervalLocScale(self.kwargs)

class powerlognorm(ContinuousReal):
    """A power log-normal continuous random variable."""
    dist = scipy.stats.powerlognorm
    def get_domain(self): return RealsPosLoc(self.kwargs)

class powernorm(ContinuousReal):
    """A power normal continuous random variable."""
    dist = scipy.stats.powernorm
    def get_domain(self): return RealsPosLoc(self.kwargs)

class rdist(ContinuousReal):
    """An R-distributed (symmetric beta) continuous random variable."""
    dist = scipy.stats.rdist
    def get_domain(self): return Interval(-1, 1)

class rayleigh(ContinuousReal):
    """A Rayleigh continuous random variable."""
    dist = scipy.stats.rayleigh
    def get_domain(self): return RealsPosLoc(self.kwargs)

class rice(ContinuousReal):
    """A Rice continuous random variable."""
    dist = scipy.stats.rice
    def get_domain(self): return RealsPosLoc(self.kwargs)

class recipinvgauss(ContinuousReal):
    """A reciprocal inverse Gaussian continuous random variable."""
    dist = scipy.stats.recipinvgauss
    def get_domain(self): return RealsPosLoc(self.kwargs)

class semicircular(ContinuousReal):
    """A semicircular continuous random variable."""
    dist = scipy.stats.semicircular
    def get_domain(self): return Interval(-1, 1)

class skewnorm(ContinuousReal):
    """A skew-normal random variable."""
    dist = scipy.stats.skewnorm
    def get_domain(self): return Reals

class t(ContinuousReal):
    """A Student’s t continuous random variable."""
    dist = scipy.stats.t
    def get_domain(self): return Reals

class trapz(ContinuousReal):
    """A trapezoidal continuous random variable."""
    dist = scipy.stats.trapz
    def get_domain(self):
        loc = self.kwargs.get('loc', 0)
        scale = self.kwargs.get('scale', 1)
        return Interval(loc, loc+scale)

class triang(ContinuousReal):
    """A triangular continuous random variable."""
    dist = scipy.stats.triang
    def get_domain(self):
        loc = self.kwargs.get('loc', 0)
        scale = self.kwargs.get('scale', 1)
        return Interval(loc, loc+scale)

class truncexpon(ContinuousReal):
    """A truncated exponential continuous random variable."""
    dist = scipy.stats.truncexpon
    def get_domain(self): return Interval(0, self.kwargs['b'])

class truncnorm(ContinuousReal):
    """A truncated normal continuous random variable."""
    dist = scipy.stats.truncnorm
    def get_domain(self): return Interval(self.kwargs['a'], self.kwargs['b'])

class tukeylambda(ContinuousReal):
    """A Tukey-Lamdba continuous random variable."""
    dist = scipy.stats.tukeylambda
    def get_domain(self): return RealsPosLoc(self.kwargs)

class uniform(ContinuousReal):
    """A uniform continuous random variable."""
    dist = scipy.stats.uniform
    def get_domain(self):
        loc = self.kwargs.get('loc', 0)
        scale = self.kwargs.get('scale', 1)
        return Interval(loc, loc + scale)

class vonmises(ContinuousReal):
    """A Von Mises continuous random variable."""
    dist = scipy.stats.vonmises
    def get_domain(self): return Interval(-sympy.pi, sympy.pi)

class vonmises_line(ContinuousReal):
    """A Von Mises continuous random variable."""
    dist = scipy.stats.vonmises_line
    def get_domain(self): return Interval(-sympy.pi, sympy.pi)

class wald(ContinuousReal):
    """A Wald continuous random variable."""
    dist = scipy.stats.wald
    def get_domain(self): return RealsPosLoc(self.kwargs)

class weibull_min(ContinuousReal):
    """Weibull minimum continuous random variable."""
    dist = scipy.stats.weibull_min
    def get_domain(self): return RealsPosLoc(self.kwargs)

class weibull_max(ContinuousReal):
    """Weibull maximum continuous random variable."""
    dist = scipy.stats.weibull_max
    def get_domain(self): return RealsNeg

class wrapcauchy(ContinuousReal):
    """A wrapped Cauchy continuous random variable."""
    dist = scipy.stats.wrapcauchy
    def get_domain(self): return Interval(0, 2*sympy.pi)

# ==============================================================================
# DiscreteReal

from .sets import Integers
from .sets import IntegersPos
from .sets import IntegersPos0
from .sets import Range
from .spn import DiscreteLeaf

class DiscreteReal(RealDistribution):
    constructor = DiscreteLeaf

class bernoulli(DiscreteReal):
    """A Bernoulli discrete random variable."""
    dist = scipy.stats.bernoulli
    def get_domain(self): return Range(0, 1)

class betabinom(DiscreteReal):
    """A beta-binomial discrete random variable."""
    dist = scipy.stats.betabinom
    def get_domain(self): return Range(0, self.kwargs['n'])

class binom(DiscreteReal):
    """A binomial discrete random variable."""
    dist = scipy.stats.binom
    def get_domain(self): return Range(0, self.kwargs['n'])

class boltzmann(DiscreteReal):
    """A Boltzmann (Truncated Discrete Exponential) random variable."""
    dist = scipy.stats.boltzmann
    def get_domain(self): return Range(0, self.kwargs['N'])

class dlaplace(DiscreteReal):
    """A Laplacian discrete random variable."""
    dist = scipy.stats.dlaplace
    def get_domain(self): return Integers

class geom(DiscreteReal):
    """A geometric discrete random variable."""
    dist = scipy.stats.geom
    def get_domain(self): return Integers

class hypergeom(DiscreteReal):
    """A hypergeometric discrete random variable."""
    dist = scipy.stats.hypergeom
    def get_domain(self):
        low = max(0, self.kwargs['N'], self.kwargs['N']-self.kwargs['M']+self.kwargs['n'])
        high = min(self.kwargs['n'], self.kwargs['N'])
        return Range(low, high)

class logser(DiscreteReal):
    """A Logarithmic (Log-Series, Series) discrete random variable."""
    dist = scipy.stats.logser
    def get_domain(self): return IntegersPos

class nbinom(DiscreteReal):
    """A negative binomial discrete random variable."""
    dist = scipy.stats.nbinom
    def get_domain(self): return IntegersPos0

class planck(DiscreteReal):
    """A Planck discrete exponential random variable."""
    dist = scipy.stats.planck
    def get_domain(self): return IntegersPos0

class poisson(DiscreteReal):
    """A Poisson discrete random variable."""
    dist = scipy.stats.poisson
    def get_domain(self): return IntegersPos0

class randint(DiscreteReal):
    """A uniform discrete random variable."""
    dist = scipy.stats.randint
    def get_domain(self): return Interval.Ropen(self.kwargs['low'], self.kwargs['high'])

class skellam(DiscreteReal):
    """A Skellam discrete random variable."""
    dist = scipy.stats.skellam
    def get_domain(self): return Integers

class zipf(DiscreteReal):
    """A Zipf discrete random variable."""
    dist = scipy.stats.zipf
    def get_domain(self): return IntegersPos

class yulesimon(DiscreteReal):
    """A Yule-Simon discrete random variable."""
    dist = scipy.stats.yulesimon
    def get_domain(self): return IntegersPos

class atomic(randint):
    """An atomic discrete random variable."""
    def __init__(self, *args, **kwargs):
        loc = kwargs.pop('loc')
        kwargs['low'] = loc
        kwargs['high'] = loc + 1
        super().__init__(*args, **kwargs)

class rv_discrete(DiscreteReal):
    """A general discrete random variable."""
    dist = lambda self, **kwargs: scipy.stats.rv_discrete(**kwargs).freeze()
    def get_domain(self):
        atoms = self.kwargs['values'][0]
        return Range(min(atoms), max(atoms))

class uniformd(rv_discrete):
    def __init__(self, *args, **kwargs):
        xk = tuple(kwargs.pop('values'))
        pk = [1./len(xk)] * len(xk)
        kwargs['values'] = (xk, pk)
        super().__init__(*args, **kwargs)

class discrete(rv_discrete):
    def __init__(self, *args, **kwargs):
        assert len(args) == 1
        assert not kwargs
        values = dict(args[0])
        xk = tuple(values.keys())
        pk = tuple(values.values())
        kwargs['values'] = (xk, pk)
        super().__init__(**kwargs)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import os

import pytest

from sppl.compilers.ast_to_spn import Id
from sppl.compilers.ast_to_spn import IfElse
from sppl.compilers.ast_to_spn import Otherwise
from sppl.compilers.ast_to_spn import Sample
from sppl.compilers.ast_to_spn import Sequence
from sppl.compilers.ast_to_spn import Transform
from sppl.distributions import bernoulli
from sppl.distributions import choice
from sppl.render import render_nested_lists
from sppl.render import render_nested_lists_concise

def get_model():
    Y = Id('Y')
    X = Id('X')
    Z = Id('Z')
    command = Sequence(
        Sample(Y,     choice({'0': .2, '1': .2, '2': .2, '3': .2, '4': .2})),
        Sample(Z,     bernoulli(p=0.1)),
        IfElse(
            Y << {str(0)} | Z << {0},  Sample(X, bernoulli(p=1/(0+1))),
            Otherwise,                 Transform(X, Z**2 + Z)))
    return command.interpret()

def test_render_lists_crash():
    model = get_model()
    render_nested_lists_concise(model)
    render_nested_lists(model)

def test_render_graphviz_crash__magics_():
    pytest.importorskip('graphviz')
    pytest.importorskip('pygraphviz')
    pytest.importorskip('networkx')

    from sppl.magics.render import render_networkx_graph
    from sppl.magics.render import render_graphviz

    model = get_model()
    render_networkx_graph(model)
    for fname in [None, '/tmp/spn.test.render']:
        for e in ['pdf', 'png', None]:
            render_graphviz(model, fname, ext=e)
            if fname is not None:
                assert not os.path.exists(fname)
                for ext in ['dot', e]:
                    f = '%s.%s' % (fname, ext,)
                    if e is not None:
                        os.path.exists(f)
                        os.unlink(f)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from collections import Callable
from functools import reduce
from itertools import chain
from itertools import product
from math import isinf

import sympy

from sympy.abc import X as symX

from sympy.calculus.util import function_range
from sympy.calculus.util import limit

from .math_util import isinf_neg
from .math_util import isinf_pos

from .poly import solve_poly_equality
from .poly import solve_poly_inequality

from .sets import EmptySet
from .sets import ExtReals
from .sets import ExtRealsPos
from .sets import FiniteNominal
from .sets import FiniteReal
from .sets import Interval
from .sets import Reals
from .sets import Strings
from .sets import Union
from .sets import convert_sympy
from .sets import make_intersection
from .sets import make_union
from .sets import oo

from .sym_util import get_union
from .sym_util import sympify_number

# ==============================================================================
# Transform base class.

class Transform():
    subexpr = None
    symbols = None

    def get_symbols(self):
        return self.symbols
    def domain(self):
        raise NotImplementedError()
    def range(self):
        raise NotImplementedError()

    def substitute(self, env):
        # TODO: Reject following cases, which result in infinite loop:
        # (1) env[X] = f(X)                 // check for self-reference
        # (2) env[X] = f(Z); env[Z] = g(X)  // detect cycle in digraph
        event_prime = self
        while expr_in_env(event_prime, env):
            event_prime = event_prime.subs(env)
        return event_prime
    def subs(self, env):
        raise NotImplementedError()

    def evaluate(self, assignment):
        raise NotImplementedError()
    def ffwd(self, x):
        raise NotImplementedError()
    def finv(self, y):
        raise NotImplementedError()

    def invert(self, ys):
        intersection = self.range() & ys
        if intersection is EmptySet:
            return EmptySet
        if isinstance(intersection, FiniteReal):
            return self.invert_finite(intersection)
        if isinstance(intersection, Interval):
            return self.invert_interval(intersection)
        if isinstance(intersection, Union):
            xs_list = [self.invert(ys_i) for ys_i in intersection.args]
            return make_union(*xs_list)
        assert False, 'Unknown intersection: %s' % (intersection,)
    def invert_finite(self, ys):
        raise NotImplementedError()
    def invert_interval(self, ys):
        # Should be called on subset of range.
        raise NotImplementedError()

    # Addition.
    def __add__number(self, x):
        poly_self = polyify(self)
        x_val = sympify_number(x)
        coeffs_new = list(poly_self.coeffs)
        coeffs_new[0] += x_val
        return Poly(poly_self.subexpr, coeffs_new)
    def __add__poly(self, x):
        if not isinstance(x, Transform):
            raise TypeError
        poly_self = polyify(self)
        poly_x = polyify(x)
        if poly_x.subexpr != poly_self.subexpr:
            raise ValueError('Incompatible subexpressions in "%s + %s"'
                % (str(self), x))
        sym_poly_a = sympy.Poly(poly_self.symexpr)
        sym_poly_b = sympy.Poly(poly_x.symexpr)
        sym_poly_c = sym_poly_a + sym_poly_b
        coeffs = sym_poly_c.all_coeffs()[::-1]
        return Poly(poly_self.subexpr, coeffs)
    def __add__(self, x):
        # Try to add x as a number.
        try:
            return self.__add__number(x)
        except TypeError:
            pass
        # Try to add x as a polynomial.
        try:
            return self.__add__poly(x)
        except TypeError:
            pass
        # Failed.
        return NotImplemented
    def __radd__(self, x):
        # Prevent infinite recursion from polymorphism.
        if not isinstance(x, Transform):
            return self + x
        return NotImplemented

    # Multiplication.
    def __mul__number(self, x):
        poly_self = polyify(self)
        x_val = sympify_number(x)
        coeffs = [x_val*c for c in poly_self.coeffs]
        return Poly(poly_self.subexpr, coeffs)
    def __mul__event(self, x):
        if not isinstance(x, Event):
            raise TypeError
        return Piecewise((self,), (x,))
    def __mul__poly(self, x):
        if not isinstance(x, Transform):
            raise TypeError
        poly_self = polyify(self)
        poly_x = polyify(x)
        if poly_x.subexpr != poly_self.subexpr:
            raise ValueError('Incompatible subexpressions in "%s * %s"'
                % (str(self), x))
        sym_poly_a = sympy.Poly(poly_self.symexpr, symX)
        sym_poly_b = sympy.Poly(poly_x.symexpr, symX)
        sym_poly_c = sym_poly_a * sym_poly_b
        coeffs = sym_poly_c.all_coeffs()[::-1]
        return Poly(poly_self.subexpr, coeffs)
    def __mul__(self, x):
        # Try to multiply x as a number.
        try:
            return self.__mul__number(x)
        except TypeError:
            pass
        # Try to multiply x as an event.
        try:
            return self.__mul__event(x)
        except TypeError:
            pass
        # Try to multiply x as a polynomial.
        try:
            return self.__mul__poly(x)
        except TypeError:
            pass
        # Failed.
        return NotImplemented
    def __rmul__(self, x):
        # Prevent infinite recursion from polymorphism.
        if not isinstance(x, Transform):
            return self * x
        return NotImplemented

    # Division by x.
    def __truediv__number(self, x):
        x_val = sympify_number(x)
        return sympy.Rational(1, x_val) * self
    def __truediv__(self, x):
        # Try to divide by x as a number.
        try:
            return self.__truediv__number(x)
        except TypeError:
            pass
        # Failed.
        return NotImplemented

    # Division by self.
    def __rtruediv__number(self, x):
        x_val = sympify_number(x)
        return Reciprocal(self) if (x_val == 1) else x_val * Reciprocal(self)
    def __rtruediv__(self, x):
        # Try to divide by x as a number.
        try:
            return self.__rtruediv__number(x)
        except TypeError:
            pass
        # Failed.
        return NotImplemented

    # Subtraction.
    def __sub__(self, x):
        return self + (-1 * x)
    def __rsub__(self, x):
        return self - x
    # Negation.
    def __neg__(self):
        return -1 * self
    # Absolute value.
    def __abs__(self):
        return Abs(self)

    # Power (self**x).
    def __pow__integer(self, x):
        if 0 < x:
            return Pow(self, x)
        if x < 0:
            return 1 / Pow(self, -x)
        raise ValueError('Cannot raise %s to %s' % (str(self), x))
    def __pow__rational(self, x):
        (numer, denom) = x.as_numer_denom()
        if numer == 1:
            return Radical(self, denom)
        if numer == -1:
            return 1 / Radical(self, denom)
        if denom == 1:
            return self.__pow__integer(numer)
        # TODO: Consider default choice x**(a/b) = (x**(a))**(1/b)
        raise ValueError('Cannot raise %s to %s' % (str(self), x))
    def __pow__number(self, x):
        x_val = sympify_number(x)
        if isinstance(x_val, (sympy.Integer, int)):
            return self.__pow__integer(x_val)
        if isinstance(x_val, sympy.Rational):
            return self.__pow__rational(x_val)
        # TODO: Convert floating-point power to rational.
        # if isinstance(x_val, sympy.Float):
        #     x_val_rat = sympy.Rational(x_val)
        #     return self.__pow__rational(x_val_rat)
        raise ValueError(
            'Cannot raise %s to irrational or floating-point power %s'
            % (str(self), x))
    def __pow__tuple(self, x):
        if not isinstance(x, tuple):
            raise TypeError
        numer = sympify_number(x[0])
        denom = sympify_number(x[1])
        return self.__pow__rational(sympy.Rational(numer, denom))
    def __pow__(self, x):
        # Try to raise to x as a number.
        try:
            return self.__pow__number(x)
        except TypeError:
            pass
        # Try to raise to x as a tuple.
        try:
            return self.__pow__tuple(x)
        except TypeError:
            pass
        # Failed.
        return NotImplemented

    # Exponentiation (x**self)
    def __rpow__number(self, x):
        x_val = sympify_number(x)
        if x_val <= 0:
            raise ValueError('Base must be positive, not %s' % (x,))
        return Exponential(self, x_val)
    def __rpow__(self, x):
        try:
            return self.__rpow__number(x)
        except TypeError:
            pass
        # Failed to exponentiate.
        return NotImplemented

    # Comparison.
    # TODO: Considering (X < Y) to mean (X - Y < 0)
    # Complication: X and Y may not have a natural ordering.
    def __le__(self, x):
        # self <= x
        try:
            x_val = sympify_number(x)
            interval = Interval(-oo, x_val)
            return EventInterval(self, interval)
        except TypeError:
            return NotImplemented
    def __lt__(self, x):
        # self < x
        try:
            x_val = sympify_number(x)
            interval = Interval(-oo, x_val, right_open=True)
            return EventInterval(self, interval)
        except TypeError:
            return NotImplemented
    def __ge__(self, x):
        # self >= x
        try:
            x_val = sympify_number(x)
            interval = Interval(x_val, oo)
            return EventInterval(self, interval)
        except TypeError:
            return NotImplemented
    def __gt__(self, x):
        # self > x
        try:
            x_val = sympify_number(x)
            interval = Interval(x_val, oo, left_open=True)
            return EventInterval(self, interval)
        except TypeError:
            return NotImplemented

    # Containment
    def __lshift__(self, x):
        if x is EmptySet:
            return EventFiniteReal(self, x)
        if isinstance(x, FiniteReal):
            return EventFiniteReal(self, x)
        if isinstance(x, FiniteNominal):
            return EventFiniteNominal(self, x)
        if isinstance(x, Interval):
            return EventInterval(self, x)
        if isinstance(x, (list, tuple, set, frozenset)):
            values = list(x)
            values_num = [v for v in values if not isinstance(v, str)]
            values_str = [v for v in values if isinstance(v, str)]
            if values_num and values_str:
                event_num = EventFiniteReal(self, FiniteReal(*values_num))
                event_str = EventFiniteNominal(self, FiniteNominal(*values_str))
                return EventOr([event_num, event_str])
            if values_num:
                return EventFiniteReal(self, FiniteReal(*values_num))
            if values_str:
                return EventFiniteNominal(self, FiniteNominal(*values_str))
            return EventFiniteReal(self, EmptySet)
        if isinstance(x, Union):
            events = [self << y for y in x.args]
            return reduce(lambda state, event: state | event, events)
        return NotImplemented

# ==============================================================================
# Injective (one-to-one) Transforms.

class Injective(Transform):
    def invert_finite(self, ys):
        ys_prime = make_union(*[self.finv(y) for y in ys])
        return self.subexpr.invert(ys_prime)
    def invert_interval(self, ys):
        assert isinstance(ys, Interval)
        (a, b) = (ys.left, ys.right)
        a_prime = next(iter(self.finv(a)))
        b_prime = next(iter(self.finv(b)))
        ys_prime = transform_interval(ys, a_prime, b_prime)
        return self.subexpr.invert(ys_prime)

class Identity(Injective):
    def __init__(self, token):
        assert isinstance(token, str)
        self.subexpr = self
        self.token = token
        self.hash = hash((self.__class__, self.token))
        self.symbols = frozenset({self})
    def domain(self):
        return ExtReals
    def range(self):
        return ExtReals
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        return env[self]
    def evaluate(self, assignment):
        if self not in assignment:
            raise ValueError('Cannot evaluate %s on %s' % (str(self), assignment))
        return self.ffwd(assignment[self])
    def ffwd(self, x):
        # assert x in self.domain()
        return x
    def finv(self, y):
        if not y in self.range():
            return EmptySet
        return FiniteReal(y)
    def invert_finite(self, ys):
        return ys
    def invert_interval(self, ys):
        return ys
    def __eq__(self, x):
        return isinstance(x, Identity) and self.token == x.token
    def __repr__(self):
        return 'Identity(%s)' % (repr(self.token),)
    def __str__(self):
        return self.token
    def __hash__(self):
        return self.hash
    def __rshift__(self, f):
        if isinstance(f, Callable):
            return f(self)
        # if isinstance(f, dict):
        #     from .spn import NominalLeaf
        #     return NominalLeaf(self, f)
        return NotImplemented

class Radical(Injective):
    def __init__(self, subexpr, degree):
        assert degree != 0
        self.subexpr = make_subexpr(subexpr, self)
        self.degree = degree
    def domain(self):
        return ExtRealsPos
    def range(self):
        return ExtRealsPos
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        subexpr_prime = self.subexpr.subs(env)
        return Radical(subexpr_prime, self.degree)
    def evaluate(self, assignment):
        x = self.subexpr.evaluate(assignment)
        return self.ffwd(x)
    def ffwd(self, x):
        assert x in self.domain()
        return sympy.Pow(x, sympy.Rational(1, self.degree))
    def finv(self, y):
        if y not in self.range():
            return EmptySet
        if isinf_pos(y):
            return FiniteReal(oo)
        return FiniteReal(sympy.Pow(y, sympy.Rational(self.degree, 1)))
    def __eq__(self, x):
        return isinstance(x, Radical) \
            and self.subexpr == x.subexpr \
            and self.degree == x.degree
    def __repr__(self):
        return 'Radical(%s, degree=%s)' \
            % (repr(self.subexpr), repr(self.degree),)
    def __str__(self):
        return '(%s)**(1, %d)' % (str(self.subexpr), self.degree)
    def __hash__(self):
        x = (self.__class__, self.subexpr, self.degree)
        return hash(x)

class Exponential(Injective):
    def __init__(self, subexpr, base):
        assert base > 0
        self.subexpr = make_subexpr(subexpr, self)
        self.base = base
    def domain(self):
        return ExtReals
    def range(self):
        return ExtRealsPos
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        subexpr_prime = self.subexpr.subs(env)
        return Exponential(subexpr_prime, self.base)
    def evaluate(self, assignment):
        x = self.subexpr.evaluate(assignment)
        return self.ffwd(x)
    def ffwd(self, x):
        assert x in self.domain()
        return sympy.Pow(self.base, x)
    def finv(self, y):
        if not y in self.range():
            return EmptySet
        if isinf_pos(y):
            return FiniteReal(oo)
        if y <= 0:
            return FiniteReal(-oo)
        return FiniteReal(sympy.log(y, self.base))
    def __eq__(self, x):
        return isinstance(x, Exponential) \
            and self.subexpr == x.subexpr \
            and self.base == x.base
    def __repr__(self):
        return 'Exponential(%s, base=%s)' \
            % (repr(self.subexpr), repr(self.base),)
    def __str__(self):
        if self.base == sympy.E:
            return 'exp(%s)' % (str(self.subexpr),)
        return '%s**(%s)' % (self.base, str(self.subexpr))
    def __hash__(self):
        x = (self.__class__, self.subexpr, self.base)
        return hash(x)

class Logarithm(Injective):
    def __init__(self, subexpr, base):
        assert base > 1
        self.subexpr = make_subexpr(subexpr, self)
        self.base = base
    def domain(self):
        return ExtRealsPos
    def range(self):
        return ExtReals
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        subexpr_prime = self.subexpr.subs(env)
        return Logarithm(subexpr_prime, self.base)
    def evaluate(self, assignment):
        x = self.subexpr.evaluate(assignment)
        return self.ffwd(x)
    def ffwd(self, x):
        assert x in self.domain()
        return {sympy.log(x, self.base) if x > 0 else -oo}
    def finv(self, y):
        if not y in self.range():
            return EmptySet
        if isinf_pos(y):
            return FiniteReal(oo)
        return FiniteReal(sympy.Pow(self.base, y))
    def __eq__(self, x):
        return isinstance(x, Logarithm) \
            and self.subexpr == x.subexpr \
            and self.base == x.base
    def __repr__(self):
        return 'Logarithm(%s, base=%s)' \
            % (repr(self.subexpr), repr(self.base))
    def __str__(self):
        if self.base == sympy.E:
            return 'ln(%s)' % (str(self.subexpr),)
        if self.base == 2:
            return 'log2(%s)' % (str(self.subexpr),)
        return 'log(%s; %s)' % (str(self.subexpr), self.base)
    def __hash__(self):
        x = (self.__class__, self.subexpr, self.base)
        return hash(x)

# ==============================================================================
# Non-injective real-valued Transforms.

class Abs(Transform):
    def __init__(self, subexpr):
        self.subexpr = make_subexpr(subexpr, self)
    def domain(self):
        return ExtReals
    def range(self):
        return ExtRealsPos
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        subexpr_prime = self.subexpr.subs(env)
        return Abs(subexpr_prime)
    def evaluate(self, assignment):
        x = self.subexpr.evaluate(assignment)
        return self.ffwd(x)
    def ffwd(self, x):
        assert x in self.domain()
        return x if x > 0 else -x
    def finv(self, y):
        if not y in self.range():
            return EmptySet
        return FiniteReal(y, -y)
    def invert_finite(self, ys):
        ys_prime = make_union(*[self.finv(y) for y in ys])
        return self.subexpr.invert(ys_prime)
    def invert_interval(self, ys):
        assert isinstance(ys, Interval)
        (a, b) = (ys.left, ys.right)
        ys_pos = transform_interval(ys, a, b)
        ys_neg = transform_interval(ys, -b, -a, flip=True)
        ys_prime = ys_pos | ys_neg
        return self.subexpr.invert(ys_prime)
    def __eq__(self, x):
        return isinstance(x, Abs) and self.subexpr == x.subexpr
    def __repr__(self):
        return 'Abs(%s)' % (repr(self.subexpr))
    def __str__(self):
        return '|%s|' % (str(self.subexpr),)
    def __hash__(self):
        x = (self.__class__, self.subexpr)
        return hash(x)
    def __abs__(self):
        return Abs(self.subexpr)

class Reciprocal(Transform):
    def __init__(self, subexpr):
        self.subexpr = make_subexpr(subexpr, self)
    def domain(self):
        # return ExtReals - sympy.FiniteSet(0)
        return FiniteReal(oo, -oo) | Interval.Ropen(-oo, 0) | Interval.Lopen(0, oo)
    def range(self):
        return Interval.Ropen(-oo, 0) | Interval.Lopen(0, oo)
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        subexpr_prime = self.subexpr.subs(env)
        return Reciprocal(subexpr_prime)
    def evaluate(self, assignment):
        x = self.subexpr.evaluate(assignment)
        return self.ffwd(x)
    def ffwd(self, x):
        assert x in self.domain()
        return 0 if isinf(x) else sympy.Pow(x, -1)
    def finv(self, y):
        if y not in self.range():
            return EmptySet
        if y == 0:
            return FiniteReal(-oo, oo)
        return FiniteReal(sympy.Pow(y, -1))
    def invert_finite(self, ys):
        ys_prime = make_union(*[self.finv(y) for y in ys])
        return self.subexpr.invert(ys_prime)
    def invert_interval(self, ys):
        (a, b) = (ys.left, ys.right)
        if (0 <= a < b):
            assert 0 < a or ys.left_open
            a_inv = sympy.Pow(a, -1) if 0 < a else oo
            b_inv = sympy.Pow(b, -1) if (not isinf(b)) else 0
            ys_prime = transform_interval(ys, b_inv, a_inv, flip=True)
            return self.subexpr.invert(ys_prime)
        if (a < b <= 0):
            assert b < 0 or ys.right_open
            a_inv = sympy.Pow(a, -1) if (not isinf(a)) else 0
            b_inv = sympy.Pow(b, -1) if b < 0 else -oo
            ys_prime = transform_interval(ys, b_inv, a_inv, flip=True)
            return self.subexpr.invert(ys_prime)
        assert False, 'Impossible Reciprocal interval: %s ' % (ys,)
    def __eq__(self, x):
        return isinstance(x, Reciprocal) \
            and self.subexpr == x.subexpr
    def __repr__(self):
        return 'Reciprocal(%s)' % (repr(self.subexpr),)
    def __str__(self):
        return '(1/%s)' % (str(self.subexpr),)
    def __hash__(self):
        x = (self.__class__, self.subexpr)
        return hash(x)

class Poly(Transform):
    def __init__(self, subexpr, coeffs):
        assert len(coeffs) > 1
        self.subexpr = make_subexpr(subexpr, self)
        self.coeffs = tuple(coeffs)
        self.degree = len(coeffs) - 1
        self.symexpr = make_sympy_polynomial(coeffs)
    def domain(self):
        return ExtReals
    def range(self):
        result = function_range(self.symexpr, symX, sympy.Reals)
        if isinstance(result, sympy.FiniteSet):
            return FiniteReal(*[float(x) for x in result.args])
        pos_inf = FiniteReal(oo) if result.right == oo else EmptySet
        neg_inf = FiniteReal(-oo) if result.left == -oo else EmptySet
        return make_union(convert_sympy(result), pos_inf, neg_inf)
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        subexpr_prime = self.subexpr.subs(env)
        return Poly(subexpr_prime, self.coeffs)
    def evaluate(self, assignment):
        x = self.subexpr.evaluate(assignment)
        return self.ffwd(x)
    def ffwd(self, x):
        assert x in self.domain()
        return self.symexpr.subs(symX, x) \
            if not isinf(x) else limit(self.symexpr, symX, x)
    def finv(self, y):
        if not y in self.range():
            return EmptySet
        return solve_poly_equality(self.symexpr, y)
    def invert_finite(self, ys):
        ys_prime = make_union(*[self.finv(y) for y in ys])
        return self.subexpr.invert(ys_prime)
    def invert_interval(self, ys):
        assert isinstance(ys, Interval)
        (a, b) = (ys.left, ys.right)
        (lo, ro) = (not ys.left_open, ys.right_open)
        ys_prime_a = solve_poly_inequality(self.symexpr, a, lo, extended=False)
        ys_prime_b = solve_poly_inequality(self.symexpr, b, ro, extended=False)
        ys_prime = ys_prime_b & (~ys_prime_a)
        return self.subexpr.invert(ys_prime)
    def __eq__(self, x):
        return isinstance(x, Poly) \
            and self.subexpr == x.subexpr \
            and self.coeffs == x.coeffs
    def __neg__(self):
        return Poly(self.subexpr, [-c for c in self.coeffs])
    def __repr__(self):
        return 'Poly(%s, coeffs=%s)' \
            % (repr(self.subexpr), repr(self.coeffs))
    def __str__(self):
        ss = str(self.subexpr)
        def term_to_str(i, c):
            if c == 0:
                return ''
            if i == 0:
                return str(c)
            if i == 1:
                return '%s(%s)' % (str(c), ss) if c != 1 else ss
            if i < len(self.coeffs):
                return '%s*(%s)**%d' % (str(c), ss, i) if c != 1 \
                    else '(%s)**%d' % (ss, i)
            assert False
        terms = [term_to_str(i, c)  for i, c in enumerate(self.coeffs)]
        return ' + '.join([t for t in terms if t])
    def __hash__(self):
        x = (self.__class__, self.subexpr, self.coeffs)
        return hash(x)

# ==============================================================================
# Non-injective Piecewise Transform.

class Piecewise(Transform):
    def __init__(self, subexprs, events):
        self.subexprs = [make_subexpr(subexpr) for subexpr in subexprs]
        self.events = [make_event(event) for event in events]
        self.symbols = get_piecewise_symbol(self.subexprs, self.events)
        self.domains = get_piecewise_domains(self.events)
    def domain(self):
        return make_union(*self.domains)
    def range(self):
        ranges = [subexpr.range() for subexpr in self.subexprs]
        return make_union(*ranges)
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        # TODO: Check whether this logic is correct.
        subexprs_prime = [subexpr.subs(env) for subexpr in self.subexprs]
        events_prime = [event.subs(env) for event in self.events]
        return Piecewise(subexprs_prime, events_prime)
    def evaluate(self, assignment):
        raise NotImplementedError()
    def ffwd(self, x):
        index = next(i for i, domain in enumerate(self.domains) if x in domain)
        return self.subexprs[index].ffwd(x)
    def finv(self, y):
        xs_list = get_piecewise_inverse(lambda subexpr: subexpr.finv(y),
            self.subexprs, self.domains)
        return make_union(*xs_list)
    def invert_finite(self, ys):
        xs_list = get_piecewise_inverse(lambda subexpr: subexpr.invert(ys),
            self.subexprs, self.domains)
        return make_union(*xs_list)
    def invert_interval(self, ys):
        xs_list = get_piecewise_inverse(lambda subexpr: subexpr.invert(ys),
            self.subexprs, self.domains)
        return make_union(*xs_list)
    def __add__(self, x):
        if isinstance(x, Piecewise):
            subexprs = self.subexprs + x.subexprs
            events = self.events + x.events
            return Piecewise(subexprs, events)
        return super().__add__(x)
    def __eq__(self, x):
        return isinstance(x, Piecewise) \
            and self.subexprs == x.subexprs \
            and self.events == x.events
    def __repr__(self):
        return 'Piecewise(%s, events=%s)' \
            % (repr(self.subexprs), repr(self.events))
    def __str__(self):
        strings = [
            '(%s) * Indicator[%s]' % (str(subexpr), str(event))
            for subexpr, event in zip(self.subexprs, self.events)
        ]
        return ' + '.join(strings)
    def __hash__(self):
        x = (self.__class__, self.subexprs, self.events)
        return hash(x)

def get_piecewise_symbol(subexprs, events):
    if len(subexprs) != len(events):
        raise ValueError('Piecewise requires same no. of subexprs and events.')
    symbols_subexprs = get_union([subexpr.get_symbols() for subexpr in subexprs])
    if len(symbols_subexprs) > 1:
        raise ValueError('Piecewise cannot have multi-symbol subexpressions.')
    symbols_events = get_union([event.get_symbols() for event in events])
    if len(symbols_subexprs) > 1:
        raise ValueError('Piecewise cannot have multi-symbol events.')
    if symbols_subexprs != symbols_events:
        raise ValueError('Piecewise events and subexprs need same symbols.')
    return symbols_subexprs

def get_piecewise_domains(events):
    domains = [event.solve() for event in events]
    for i, di in enumerate(domains):
        for j, dj in enumerate(domains):
            if i == j:
                continue
            intersection = di & dj
            if intersection is not EmptySet:
                raise ValueError('Piecewise events %s and %s overlap'
                    % (di, dj))
    return domains

def get_piecewise_inverse(f_inv, subexprs, domains):
    inverses = [f_inv(subexpr) for subexpr in subexprs]
    intersections = [i & d for i, d in zip(inverses, domains)]
    return make_union(intersections)

# ==============================================================================
# Non-injective Boolean-valued Transforms.

class Event(Transform):
    def range(self):
        return FiniteReal(0, 1)

    def __mul__(self, x):
        if isinstance(x, Transform):
            return Piecewise((x,), (self,))
        return super().__mul__(x)

    # Event methods.
    def solve(self):
        if len(self.symbols) > 1:
            raise ValueError('Cannot solve multi-symbol Event.')
        return self.invert(FiniteReal(1))
    def to_dnf(self):
        dnf = self.to_dnf_list()
        simplify_event = lambda x, E: x[0] if len(x)==1 else E(x)
        events = [simplify_event(conjunction, EventAnd) for conjunction in dnf]
        return simplify_event(events, EventOr)
    def to_dnf_list(self):
        raise NotImplementedError()
    def __and__(self, event):
        # Naive implementation (no simplification):
        # return EventAnd([self, event])
        raise NotImplementedError()
    def __or__(self, event):
        raise NotImplementedError()
        # Naive implementation (no simplification):
        # return EventOr([self, event])
    def __invert__(self):
        raise NotImplementedError()
    def __xor__(self, event):
        if isinstance(event, Event):
            return (self & ~event) | (~self & event)
        return NotImplemented

class EventBasic(Event):
    subexpr = None
    values = None

    def domain(self):
        return self.subexpr.domain()
    def solve(self):
        if isinstance(self.subexpr, Identity):
            return self.values
        return super().solve()
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        subexpr_prime = self.subexpr.subs(env)
        return type(self)(subexpr_prime, self.values)
    def evaluate(self, assignment):
        x = self.subexpr.evaluate(assignment)
        return self.ffwd(x)
    def ffwd(self, x):
        return x in self.values

    # Disable < on Events.
    def __gt__(self, x):
        raise TypeError()
    def __ge__(self, x):
        raise TypeError()
    def __lt__(self, x):
        raise TypeError()
    def __le__(self, x):
        raise TypeError()

    # Event methods.
    def to_dnf_list(self):
        return [[self]]
    def __and__(self, event):
        # Simplify.
        if isinstance(event, EventBasic):
            if event.subexpr == self.subexpr:
                intersection = self.values & event.values
                if intersection is EmptySet:
                    return EventFiniteReal(self.subexpr, EmptySet)
                if isinstance(intersection, FiniteReal):
                    return EventFiniteReal(self.subexpr, intersection)
                if isinstance(intersection, FiniteNominal):
                    return EventFiniteNominal(self.subexpr, intersection)
                if isinstance(intersection, Interval):
                    return EventInterval(self.subexpr, intersection)
        # Linearize but do not simplify.
        if isinstance(event, EventAnd):
            events = (self,) + event.subexprs
            return EventAnd(events)
        if isinstance(event, (EventBasic, EventOr)):
            return EventAnd([self, event])
        return NotImplemented
    def __or__(self, event):
        # Simplify.
        if isinstance(event, EventBasic):
            if event.subexpr == self.subexpr:
                union = self.values | event.values
                if union is EmptySet:
                    return EventFiniteReal(self.subexpr, EmptySet)
                if isinstance(union, FiniteReal):
                    return EventFiniteReal(self.subexpr, union)
                if isinstance(union, FiniteNominal):
                    return EventFiniteNominal(self.subexpr, union)
                if isinstance(union, Interval):
                    return EventInterval(self.subexpr, union)
        # Linearize but do not simplify.
        if isinstance(event, EventOr):
            events = (self,) + event.subexprs
            return EventOr(events)
        if isinstance(event, (EventBasic, EventAnd)):
            return EventOr([self, event])
        return NotImplemented
    def __eq__(self, event):
        return isinstance(event, type(self)) \
            and (self.values == event.values) \
            and (self.subexpr == event.subexpr)
    def __hash__(self):
        x = (self.__class__, self.subexpr, self.values)
        return hash(x)

class EventInterval(EventBasic):
    def __init__(self, subexpr, values):
        assert isinstance(values, Interval)
        self.subexpr = make_subexpr(subexpr, self)
        self.values = values
    def finv(self, y):
        if y not in self.range():
            return EmptySet
        if y == 1:
            return self.values
        if y == 0:
            return ~self.values
    def invert_finite(self, ys):
        ys_prime = make_union(*[self.finv(y) for y in ys])
        return self.subexpr.invert(ys_prime)

    # Support chaining notation (a < X) < b
    # Overrides the behavior of < from Transform.
    def __compute_gte__(self, x, left_open):
        # x < (Y < b)
        if not isinf_neg(self.values.left):
            raise ValueError('cannot compute %s < %s' % (x, str(self)))
        xn = sympify_number(x)
        # TODO: Replace interval_tmp
        interval_tmp = sympy.Interval(
            xn, self.values.right,
            left_open=left_open,
            right_open=self.values.right_open)
        interval = convert_sympy(interval_tmp)
        if isinstance(interval, Interval):
            return EventInterval(self.subexpr, interval)
        if isinstance(interval, FiniteReal) or interval is EmptySet:
            return EventFiniteReal(self.subexpr, interval)
        assert False, 'Unknown interval: %s' % (interval,)
    def __compute_lte__(self, x, right_open):
        # (a < Y) < x
        if not isinf_pos(self.values.right):
            raise ValueError('cannot compute %s < %s' % (str(self), x))
        xn = sympify_number(x)
        # TODO: Replace interval_tmp
        interval_tmp = sympy.Interval(
            self.values.left, xn,
            left_open=self.values.left_open,
            right_open=right_open)
        interval = convert_sympy(interval_tmp)
        if isinstance(interval, Interval):
            return EventInterval(self.subexpr, interval)
        if isinstance(interval, FiniteReal) or interval is EmptySet:
            return EventFiniteReal(self.subexpr, interval)
        assert False, 'Unknown interval: %s' % (interval,)
    def __gt__(self, x):
        return self.__compute_gte__(x, True)
    def __ge__(self, x):
        return self.__compute_gte__(x, False)
    def __lt__(self, x):
        return self.__compute_lte__(x, True)
    def __le__(self, x):
        return self.__compute_lte__(x, False)
    def __invert__(self):
        values_not = ~self.values
        if values_not is EmptySet:
            return EventFiniteReal(self.subexpr, values_not)
        if isinstance(values_not, Interval):
            return EventInterval(self.subexpr, values_not)
        if isinstance(values_not, Union):
            assert len(values_not.args) == 2
            event_l = EventInterval(self.subexpr, values_not.args[0])
            event_r = EventInterval(self.subexpr, values_not.args[1])
            return EventOr([event_l, event_r])
        assert False, 'Unknown complemented interval: %s' % (values_not,)
    def __repr__(self):
        return 'EventInterval(%s, values=%s)' \
            % (repr(self.subexpr), repr(self.values))
    def __str__(self):
        sym = str(self.subexpr)
        comp_l = '<' if self.values.left_open else '<='
        (x_l, x_r) = (self.values.left, self.values.right)
        comp_r = '<' if self.values.right_open else '<='
        if isinf_neg(x_l):
            result = '%s %s %s' % (sym, comp_r, x_r)
        elif isinf_pos(x_r):
            result = '%s %s %s' % (x_l, comp_l, sym)
        else:
            # Python-style parenthesis are required for
            # rendering condition(E) in spn_to_sppl.py
            result = '(%s %s %s) %s %s' % (x_l, comp_l, sym, comp_r, x_r)
        return result

class EventFiniteReal(EventBasic):
    def __init__(self, subexpr, values):
        self.subexpr = make_subexpr(subexpr, self)
        self.values = values if values is EmptySet else FiniteReal(*values)
    def finv(self, y):
        if y not in self.range():
            return EmptySet
        if y == 1:
            return self.values
        if y == 0:
            return ~self.values
    def invert_finite(self, ys):
        ys_prime = make_union(*[self.finv(y) for y in ys])
        return self.subexpr.invert(ys_prime)

    def __repr__(self):
        return 'EventFiniteReal(%s, values=%s)' \
            % (repr(self.subexpr), repr(self.values))
    def __str__(self):
        str_items = '' if self.values is EmptySet else \
            ', '.join('%s' % (x,) for x in sorted(self.values))
        return '%s << %s' % (str(self.subexpr), '{%s}' % (str_items,))
    def __invert__(self):
        values_not = ~self.values
        if values_not == Reals:
            return EventInterval(self.subexpr, Reals)
        assert isinstance(values_not, Union)
        events = [EventInterval(self.subexpr, v) for v in values_not.args]
        return EventOr(events)

class EventFiniteNominal(EventBasic):
    def __init__(self, subexpr, values):
        assert values is EmptySet or isinstance(values, FiniteNominal)
        self.subexpr = make_subexpr(subexpr, self)
        self.values = values
        self.transformed = not isinstance(self.subexpr, Identity)
        self.complemented = self.values is not EmptySet and self.values.b

    def finv(self, y):
        if y not in self.range():
            return EmptySet
        if y == 1:
            if self.values is EmptySet:
                return EmptySet
            if self.transformed:
                if not self.complemented:
                    return EmptySet
                else:
                    return Strings
            return self.values
        if y == 0:
            if self.values is EmptySet:
                return Strings
            if self.transformed:
                if not self.complemented:
                    return Strings
                else:
                    return EmptySet
            return ~self.values

    def invert_finite(self, ys):
        if ys == FiniteReal(0):
            return self.finv(0)
        if ys == FiniteReal(1):
            return self.finv(1)
        if ys == FiniteReal(1, 2):
            return Strings

    def __repr__(self):
        return 'EventFiniteNominal(%s, values=%s)' \
            % (repr(self.subexpr), repr(self.values))
    def __str__(self):
        return '%s << %s' % (str(self.subexpr), str(self.values))
    def __invert__(self):
        values_not = ~self.values
        return EventFiniteNominal(self.subexpr, values_not)

class EventCompound(Event):
    def __init__(self, subexprs):
        assert all(isinstance(subexpr, Event) for subexpr in subexprs)
        self.subexprs = tuple([make_subexpr(event) for event in subexprs])
        self.symbols = get_union([event.get_symbols() for event in subexprs])
    def domain(self):
        if len(self.symbols) > 1:
            raise ValueError('No domain for multi-symbol Event.')
        domains = [event.domain() for event in self.subexprs]
        return make_intersection(*domains)
    def subs(self, env):
        if not expr_in_env(self, env):
            return self
        subexprs_prime = [subexpr.subs(env) for subexpr in self.subexprs]
        return type(self)(subexprs_prime)

class EventOr(EventCompound):
    def evaluate(self, assignment):
        ys = [event.evaluate(assignment) for event in self.subexprs]
        return any(ys)
    def ffwd(self, x):
        # Cannot asses on multi-symbol Event.
        ys = [event.ffwd(x) for event in self.subexprs]
        return any(ys)
    def finv(self, y):
        # Cannot invert multi-symbol Event.
        xs_list = [event.finv(y) for event in self.subexprs]
        return make_union(*xs_list)
    def invert_finite(self, ys):
        xs_list = [event.invert(ys) for event in self.subexprs]
        return make_union(*xs_list)

    # Event methods.
    def to_dnf_list(self):
        sub_dnf = [event.to_dnf_list() for event in self.subexprs]
        return list(chain.from_iterable(sub_dnf))
    def __and__(self, event):
        if isinstance(event, EventAnd):
            events = (self,) + event.subexprs
            return EventAnd(events)
        if isinstance(event, (EventBasic, EventOr)):
            events = (self, event)
            return EventAnd(events)
        return NotImplemented
    def __or__(self, event):
        if isinstance(event, EventOr):
            events = self.subexprs + event.subexprs
            return EventOr(events)
        if isinstance(event, (EventBasic, EventAnd)):
            events = self.subexprs + (event,)
            return EventOr(events)
        return NotImplemented
    def __eq__(self, event):
        return isinstance(event, EventOr) and (self.subexprs == event.subexprs)
    def __invert__(self):
        sub_events = [~event for event in self.subexprs]
        return reduce(lambda state, event: state & event, sub_events)
    def __repr__(self):
        return 'EventOr(%s)' % (repr(self.subexprs,))
    def __str__(self):
        sub_events = ['(%s)' % (str(event),) for event in self.subexprs]
        return ' | '.join(sub_events)
    def __hash__(self):
        x = (self.__class__, self.subexprs)
        return hash(x)

class EventAnd(EventCompound):
    def evaluate(self, assignment):
        ys = [event.evaluate(assignment) for event in self.subexprs]
        return all(ys)
    def ffwd(self, x):
        # Cannot asses on multi-symbol Event.
        ys = [event.ffwd(x) for event in self.subexprs]
        return all(ys)
    def finv(self, y):
        # Cannot invert multi-symbol Event.
        xs_list = [event.finv(y) for event in self.subexprs]
        return make_intersection(*xs_list)
    def invert_finite(self, ys):
        xs_list = [event.invert(ys) for event in self.subexprs]
        return make_intersection(*xs_list)

    # Event methods.
    def to_dnf_list(self):
        sub_dnf = [event.to_dnf_list() for event in self.subexprs]
        return [
            list(chain.from_iterable(cross))
            for cross in product(*sub_dnf)
        ]
    def __and__(self, event):
        if isinstance(event, EventAnd):
            events = self.subexprs + event.subexprs
            return EventAnd(events)
        if isinstance(event, (EventBasic, EventOr)):
            events = self.subexprs + (event,)
            return EventAnd(events)
        return NotImplemented
    def __or__(self, event):
        if isinstance(event, EventOr):
            events = (self,) + event.subexprs
            return EventOr(events)
        if isinstance(event, (EventBasic, EventAnd)):
            events = (self, event)
            return EventOr(events)
        return NotImplemented
    def __eq__(self, event):
        return isinstance(event, EventAnd) and (self.subexprs == event.subexprs)
    def __invert__(self):
        sub_events = [~event for event in self.subexprs]
        return reduce(lambda state, event: state | event, sub_events)
    def __repr__(self):
        return 'EventAnd(%s)' % (repr(self.subexprs,))
    def __str__(self):
        sub_events = ['(%s)' % (str(event),) for event in self.subexprs]
        return ' & '.join(sub_events)
    def __hash__(self):
        x = (self.__class__, self.subexprs)
        return hash(x)

# ==============================================================================
# Utilities.

# Some useful constructors.
Id = Identity
def Exp(subexpr):
    return Exponential(subexpr, sympy.exp(1))
def Log(subexpr):
    return Logarithm(subexpr, sympy.exp(1))
def Sqrt(subexpr):
    return Radical(subexpr, 2)
def Pow(subexpr, n):
    assert 0 <= n
    coeffs = [0]*n + [1]
    return Poly(subexpr, coeffs)

def expr_in_env(expr, env):
    return env and any(env.get(s,s) != s for s in expr.get_symbols())

def transform_interval(interval, a, b, flip=None):
    return \
        Interval(a, b, interval.left_open, interval.right_open) \
        if not flip else \
        Interval(a, b, interval.right_open, interval.left_open) \

def make_sympy_polynomial(coeffs):
    terms = [c*symX**i for (i,c) in enumerate(coeffs)]
    return sympy.Add(*terms)

def make_subexpr(subexpr, expr=None):
    if isinstance(subexpr, Transform):
        if expr is not None:
            expr.symbols = subexpr.symbols
        return subexpr
    raise TypeError('Invalid subexpr %s with type %s'
        % (subexpr, str(type(subexpr))))

def make_event(event):
    if isinstance(event, Event):
        return event
    raise TypeError('Invalid event %s with type %s'
        % (event, str(type(event))))

def polyify(expr):
    if isinstance(expr, Poly):
        return expr
    return Poly(expr, (0, 1))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

from sppl.sets import EmptySet
from sppl.sets import FiniteNominal as FN
from sppl.sets import FiniteReal as FR
from sppl.sets import Interval
from sppl.sets import Union
from sppl.sets import inf
from sppl.sets import union_intervals
from sppl.sets import union_intervals_finite

def test_FiniteNominal_in():
    with pytest.raises(Exception):
        FN()
    assert 'a' in FN('a')
    assert 'b' not in FN('a')
    assert 'b' in FN('a', b=True)
    assert 'a' not in FN('a', b=True)
    assert 'a' in FN(b=True)

def test_FiniteNominal_invert():
    assert ~(FN('a')) == FN('a', b=True)
    assert ~(FN('a', b=True)) == FN('a')
    assert ~FN(b=True) == EmptySet

def test_FiniteNominal_and():
    assert FN('a','b') & EmptySet is EmptySet
    assert FN('a','b') & FN('c') is EmptySet
    assert FN('a','b','c') & FN('a') == FN('a')
    assert FN('a','b','c') & FN(b=True) == FN('a','b','c')
    assert FN('a','b','c') & FN('a') == FN('a')
    assert FN('a','b','c', b=True) & FN('a') is EmptySet
    assert FN('a','b','c', b=True) & FN('d','a','b') == FN('d')
    assert FN('a','b','c', b=True) & FN('d') == FN('d')
    assert FN('a','b','c') & FN('a', b=True) == FN('b','c')
    assert FN('a','b','c') & FN('d','a','b', b=True) == FN('c')
    assert FN('a','b','c') & FN('d', b=True) == FN('a','b','c')
    assert FN('a','b','c', b=True) & FN('d', b=True) == FN('a','b','c','d', b=True)
    assert FN('a','b','c', b=True) & FN('a', b=True) == FN('a','b','c', b=True)
    assert FN(b=True) & FN(b=True) == FN(b=True)
    # FiniteReal
    assert FN('a') & FR(1) is EmptySet
    assert FN('a', b=True) & FR(1) is EmptySet
    # Interval
    assert FN('a') & Interval(0,1) is EmptySet

def test_FiniteNominal_or():
    # EmptySet
    assert FN('a','b') | EmptySet == FN('a','b')
    # Nominal
    assert FN('a','b') | FN('c') == FN('a','b','c')
    assert FN('a','b','c') | FN('a') == FN('a','b','c')
    assert FN('a','b','c') | FN(b=True) == FN(b=True)
    assert FN('a','b','c', b=True) | FN('a') == FN('b','c', b=True)
    assert FN('a','b','c', b=True) | FN('d','a','b') == FN('c', b=True)
    assert FN('a','b','c', b=True) | FN('d') == FN('a','b','c', b=True)
    assert FN('a','b','c') | FN('a', b=True) == FN(b=True)
    assert FN('a','b','c') | FN('d','a','b', b=True) == FN('d', b=True)
    assert FN('a','b','c') | FN('d', b=True) == FN('d', b=True)
    assert FN('a','b','c', b=True) | FN('d', b=True) == FN(b=True)
    assert FN('a','b','c', b=True) | FN('a', b=True) == FN('a', b=True)
    assert FN(b=True) | FN(b=True) == FN(b=True)
    # FiniteReal
    assert FN('a') | FR(1) == Union(FN('a'), FR(1))
    assert FN('a', b=True) | FR(1) == Union(FR(1), FN('a', b=True))
    # Interval
    assert FN('a') | Interval(0,1) == Union(FN('a'), Interval(0,1))

def test_FiniteReal_in():
    with pytest.raises(Exception):
        FR()
    assert 1 in FN(1,2)
    assert 2 not in FN(1)

def test_FiniteReal_invert():
    assert ~FR(1) == Union(
        Interval.Ropen(-inf,1),
        Interval.Lopen(1, inf))
    assert ~(FR(0,-1)) == Union(
        Interval.Ropen(-inf,-1),
        Interval.open(-1, 0),
        Interval.Lopen(0, inf))

def test_FiniteReal_and():
    assert FR(1) & FR(2) is EmptySet
    assert FR(1,2) & FR(2) == FR(2)
    assert FR(1,2) & FN('2') is EmptySet
    assert FR(1,2) & FN('2', b=True) is EmptySet
    assert FR(1,2) & Interval(0,1) == FR(1)
    assert FR(1,2) & Interval.Ropen(0,1) is EmptySet
    assert FR(0,2) & Interval(0,1) == FR(0)
    assert FR(0,2) & Interval.Lopen(0,1) is EmptySet
    assert FR(-1,1) & Interval(-10,10) == FR(-1,1)
    assert FR(-1,11) & Interval(-10,10) == FR(-1)

def test_FiniteReal_or():
    assert FR(1) | FR(2) == FR(1,2)
    assert FR(1,2) | FR(2) == FR(1,2)
    assert FR(1,2) | FN('2') == Union(FR(1,2), FN('2'))
    assert FR(1,2) | Interval(0,1) == Union(Interval(0,1), FR(2))
    assert FR(1,2) | Interval.Ropen(0,1) == Union(Interval(0,1), FR(2))
    assert FR(0,1,2) | Interval.open(0,1) == Union(Interval(0,1), FR(2))
    assert FR(0,2) | Interval.Lopen(0,1) == Union(Interval(0,1), FR(2))
    assert FR(0,2) | Interval.Lopen(2.5,10) == Union(Interval.Lopen(2.5,10), FR(0,2))
    assert FR(-1,1) | Interval(-10,10) == Interval(-10,10)
    assert FR(-1,11) | Interval(-10,10) == Union(Interval(-10, 10), FR(11))

def test_Interval_in():
    with pytest.raises(Exception):
        Interval(3, 1)
    assert 1 in Interval(0,1)
    assert 1 not in Interval.Ropen(0,1)
    assert 1 in Interval.Lopen(0,1)
    assert 0 in Interval(0,1)
    assert 0 not in Interval.Lopen(0,1)
    assert 0 in Interval.Ropen(0,1)
    assert inf not in Interval(-inf, inf)
    assert -inf not in Interval(-inf, 0)
    assert 10 in Interval(-inf, inf)

def test_Interval_invert():
    assert ~(Interval(0,1)) == Union(Interval.Ropen(-inf, 0), Interval.Lopen(1, inf))
    assert ~(Interval.open(0,1)) == Union(Interval(-inf, 0), Interval(1, inf))
    assert ~(Interval.Lopen(0,1)) == Union(Interval(-inf, 0), Interval.Lopen(1, inf))
    assert ~(Interval.Ropen(0,1)) == Union(Interval.Ropen(-inf, 0), Interval(1, inf))
    assert ~(Interval(-inf, inf)) is EmptySet
    assert ~(Interval(3, inf)) == Interval.Ropen(-inf, 3)
    assert ~(Interval.open(3, inf)) == Interval(-inf, 3)
    assert ~(Interval.Lopen(3, inf)) == Interval(-inf, 3)
    assert ~(Interval.Ropen(3, inf)) == Interval.Ropen(-inf, 3)
    assert ~(Interval(-inf, 3)) == Interval.Lopen(3, inf)
    assert ~(Interval.open(-inf, 3)) == Interval(3, inf)
    assert ~(Interval.Lopen(-inf, 3)) == Interval.Lopen(3, inf)
    assert ~(Interval.Ropen(-inf, 3)) == Interval(3, inf)
    assert ~(Interval.open(-inf, inf)) is EmptySet

def test_Interval_and():
    assert Interval(0,1) & Interval(-1,1) == Interval(0,1)
    assert Interval(0,2) & Interval.open(0,1) == Interval.open(0,1)
    assert Interval.Lopen(0,1) & Interval(-1,1) == Interval.Lopen(0,1)
    assert Interval.Ropen(0,1) & Interval(-1,1) == Interval.Ropen(0,1)
    assert Interval(0,1) & Interval(1,2) == FR(1)
    assert Interval.Lopen(0,1) & Interval(1,2) == FR(1)
    assert Interval.Ropen(0,1) & Interval(1,2) is EmptySet
    assert Interval.Lopen(0,1) & Interval.Lopen(1,2) is EmptySet
    assert Interval(1,2) & Interval.Lopen(0,1) == FR(1)
    assert Interval(1,2) & Interval.open(0,1) is EmptySet
    assert Interval(0,2) & Interval.Lopen(0.5,2.5) == Interval.Lopen(0.5,2)
    assert Interval.Ropen(0,2) & Interval.Lopen(0.5,2.5) == Interval.open(0.5,2)
    assert Interval.open(0,2) & Interval(0.5,2.5) == Interval.Ropen(0.5,2)
    assert Interval.Lopen(0,2) & Interval.Ropen(0,2) == Interval.open(0,2)
    assert Interval(0,1) & Interval(2,3) is EmptySet
    assert Interval(2,3) & Interval(0,1) is EmptySet
    assert Interval.open(0,1) & Interval.open(0,1) == Interval.open(0,1)
    assert Interval.Ropen(-inf, -3) & Interval(-inf, inf) == Interval.Ropen(-inf, -3)
    assert Interval(-inf, inf) & Interval.Ropen(-inf, -3) == Interval.Ropen(-inf, -3)
    assert Interval(0, inf) & (Interval.Lopen(-5, inf)) == Interval(0, inf)
    assert Interval.Lopen(0, 1) & Interval.Ropen(0, 1) == Interval.open(0, 1)
    assert Interval.Ropen(0, 1) & Interval.Lopen(0, 1) == Interval.open(0, 1)
    assert Interval.Ropen(0, 5) & Interval.Ropen(-inf, 5) == Interval.Ropen(0, 5)

def test_Interval_or():
    assert Interval(0,1) | Interval(-1,1) == Interval(-1,1)
    assert Interval(0, 2) | Interval.open(0, 1) == Interval(0, 2)
    assert Interval.Lopen(0,1) | Interval(-1,1) == Interval(-1,1)
    assert Interval.Ropen(0,1) | Interval(-1,1) == Interval(-1,1)
    assert Interval(0,1) | Interval(1, 2) == Interval(0, 2)
    assert Interval.open(0, 1) | Interval(0,1) == Interval(0, 1)
    assert Interval(0, 1) | Interval(0,.5) == Interval(0, 1)
    assert Interval.Lopen(0, 1) | Interval(1,2) == Interval.Lopen(0, 2)
    assert Interval.Ropen(-1, 0) | Interval.Ropen(0, 1) == Interval.Ropen(-1,1)
    assert Interval.Ropen(0, 1) | Interval.Ropen(-1, 0) == Interval.Ropen(-1,1)
    assert Interval.Lopen(0, 1) | Interval.Lopen(1, 2) == Interval.Lopen(0,2)
    assert Interval.Lopen(0, 1) | Interval.Ropen(1, 2) == Interval.open(0,2)
    assert Interval.open(0, 2) | Interval(0, 1) == Interval.Ropen(0, 2)
    assert Interval.open(0, 1) | Interval.Ropen(-1, 0) == Union(Interval.open(0, 1), Interval.Ropen(-1,0))
    assert Interval(1, 2) | Interval.Ropen(0, 1) == Interval(0,2)
    assert Interval.Ropen(0, 1) | Interval(1, 2) == Interval(0,2)
    assert Interval.open(0, 1) | Interval(1, 2) == Interval.Lopen(0,2)
    assert Interval.Ropen(0, 1) | Interval.Ropen(1, 2) == Interval.Ropen(0,2)
    assert Interval(1, 2) | Interval.open(0, 1) == Interval.Lopen(0,2)
    assert Interval(1, 2) | Interval.Lopen(0, 1) == Interval.Lopen(0,2)
    assert Interval(1, 2) | Interval.Ropen(0, 1) == Interval(0,2)
    assert Interval.open(0,1) | Interval(1,2) == Interval.Lopen(0,2)
    assert Interval.Lopen(0,1) | Interval(1,2) == Interval.Lopen(0,2)
    assert Interval.Ropen(0,1) | Interval(1,2) == Interval(0,2)
    assert Interval(0,2) | Interval.open(0.5, 2.5) == Interval.Ropen(0, 2.5)
    assert Interval.open(0,2) | Interval.open(0, 2.5) == Interval.open(0, 2.5)
    assert Interval.open(0,2.5) | Interval.open(0, 2) == Interval.open(0, 2.5)
    assert Interval.open(0,1) | Interval.open(1, 2) == Union(Interval.open(0, 1), Interval.open(1,2))
    assert Interval.Ropen(0,2) | Interval.Lopen(0.5, 2.5) == Interval(0, 2.5)
    assert Interval.open(0,2) | Interval(0.5,2.5) == Interval.Lopen(0, 2.5)
    assert Interval.Lopen(0,2) | Interval.Ropen(0,2) == Interval(0,2)
    assert Interval(0,1) | Interval(2,3) == Union(Interval(0,1), Interval(2,3))
    assert Interval(2,3) | Interval.Ropen(0,1) == Union(Interval(2,3), Interval.Ropen(0,1))
    assert Interval.Lopen(0,1) | Interval(1,2) == Interval.Lopen(0,2)
    assert Interval(-10,10) | FR(-1,1) == Interval(-10,10)
    assert Interval(-10,10) | FR(-1,11) == Union(Interval(-10, 10), FR(11))
    assert Interval(-inf, -3, right_open=True) | Interval(-inf, inf) == Interval(-inf, inf)

def test_union_intervals():
    assert union_intervals([
        Interval(0,1),
        Interval(2,3),
        Interval(1,2)
    ]) == [Interval(0,3)]
    assert union_intervals([
        Interval.open(0,1),
        Interval(2,3),
        Interval(1,2)
    ]) == [Interval.Lopen(0,3)]
    assert union_intervals([
        Interval.open(0,1),
        Interval(2,3),
        Interval.Lopen(1,2)
    ]) == [Interval.open(0,1), Interval.Lopen(1,3)]
    assert union_intervals([
        Interval.open(0,1),
        Interval.Ropen(0,3),
        Interval.Lopen(1,2)
    ]) == [Interval.Ropen(0,3)]
    assert union_intervals([
        Interval.open(-2,-1),
        Interval.Ropen(0,3),
        Interval.Lopen(1,2)
    ]) == [Interval.open(-2,-1), Interval.Ropen(0,3)]

def test_union_intervals_finite():
    assert union_intervals_finite([
            Interval.open(0,1),
            Interval(2,3),
            Interval.Lopen(1,2)
        ], FR(1)) \
        == [Interval.Lopen(0, 3)]
    assert union_intervals_finite([
            Interval.open(0,1),
            Interval.open(2, 3),
            Interval.open(1,2)
        ], FR(1, 3)) \
        == [Interval.open(0, 2), Interval.Lopen(2, 3)]
    assert union_intervals_finite([
            Interval.open(0,1),
            Interval.open(1, 3),
            Interval.open(11,15)
        ], FR(1, -11, -19, 3)) \
        == [Interval.Lopen(0, 3), Interval.open(11,15), FR(-11, -19)]

def test_Union_or():
    x = Interval(0,1) | Interval(5,6) | Interval(10,11)
    assert x == Union(Interval(0,1), Interval(5,6), Interval(10,11))
    x = Interval.Ropen(0,1) | Interval.Lopen(1,2) | Interval(10,11)
    assert x == Union(Interval.Ropen(0,1), Interval.Lopen(1,2), Interval(10,11))
    x = Interval.Ropen(0,1) | Interval.Lopen(1,2) | Interval(10,11) | FR(1)
    assert x == Union(Interval(0,2), Interval(10,11))
    x = (Interval.Ropen(0,1) | Interval.Lopen(1,2)) | (Interval(10,11) | FR(1))
    assert x == Union(Interval(0,2), Interval(10,11))
    x = FR(1) | ((Interval.Ropen(0,1) | Interval.Lopen(1,2) | FR(10,13)) \
            | (Interval.Lopen(10,11) | FR(7)))
    assert x == Union(Interval(0,2), Interval(10,11), FR(13, 7))
    assert 2 in x
    assert 13 in x
    x = FN('f') | (FR(1) | FN('g', b=True))
    assert x == Union(FR(1), FN('g', b=True))
    assert 'w' in x
    assert 'g' not in x

def test_Union_and():
    x = (Interval(0,1) | FR(1)) & (FN('a'))
    assert x is EmptySet
    x = (FN('x', b=True)| Interval(0,1) | FR(1)) & (FN('a'))
    assert x == FN('a')
    x = (FN('x')| Interval(0,1) | FR(1)) & (FN('a'))
    assert x is EmptySet
    x = (FN('x')| Interval.open(0,1) | FR(7)) & ((FN('x')) | FR(.5) | Interval(.75, 1.2))
    assert x == Union(FR(.5), FN('x'), Interval.Ropen(.75, 1))
    x = (FN('x')| Interval.open(0,1) | FR(7)) & (FR(3))
    assert x is EmptySet
    x = (Interval.Lopen(-5, inf)) & (Interval(0, inf) | FR(inf))
    assert x == Interval(0, inf)
    x = (FR(1,2) | Interval.Ropen(-inf, 0)) & Interval(0, inf)
    assert x == FR(1,2)
    x = (FR(1,12) | Interval(0, 5) | Interval(7,10)) & Interval(4, 12)
    assert x == Union(Interval(4,5), Interval(7,10), FR(12))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import exp
from math import log

import numpy
import pytest

from sppl.distributions import norm
from sppl.math_util import allclose
from sppl.math_util import isinf_neg
from sppl.math_util import logdiffexp
from sppl.math_util import logsumexp
from sppl.spn import Memo
from sppl.transforms import Id

prng = numpy.random.RandomState(1)

def entropy(spn, A, memo):
    lpA1 = spn.logprob(A, memo=memo)
    lpA0 = logdiffexp(0, lpA1)
    e1 = -exp(lpA1) * lpA1 if not isinf_neg(lpA1) else 0
    e0 = -exp(lpA0) * lpA0 if not isinf_neg(lpA0) else 0
    return e1 + e0
def entropyc(spn, A, B, memo):
    lpB1 = spn.logprob(B)
    lpB0 = logdiffexp(0, lpB1)
    lp11 = spn.logprob(B & A, memo=memo)
    lp10 = spn.logprob(B & ~A, memo=memo)
    lp01 = spn.logprob(~B & A, memo=memo)
    # lp00 = self.logprob(~B & ~A, memo)
    lp00 = logdiffexp(0, logsumexp([lp11, lp10, lp01]))
    m11 = exp(lp11) * (lpB1 - lp11) if not isinf_neg(lp11) else 0
    m10 = exp(lp10) * (lpB1 - lp10) if not isinf_neg(lp10) else 0
    m01 = exp(lp01) * (lpB0 - lp01) if not isinf_neg(lp01) else 0
    m00 = exp(lp00) * (lpB0 - lp00) if not isinf_neg(lp00) else 0
    return m11 + m10 + m01 + m00

def check_mi_properties(spn, A, B, memo):
    miAB = spn.mutual_information(A, B, memo=memo)
    miAA = spn.mutual_information(A, A, memo=memo)
    miBB = spn.mutual_information(B, B, memo=memo)
    eA = entropy(spn, A, memo=memo)
    eB = entropy(spn, B, memo=memo)
    eAB = entropyc(spn, A, B, memo=memo)
    eBA = entropyc(spn, B, A, memo=memo)
    assert allclose(miAA, eA)
    assert allclose(miBB, eB)
    assert allclose(miAB, eA - eAB)
    assert allclose(miAB, eB - eBA)

@pytest.mark.parametrize('memo', [Memo(), None])
def test_mutual_information_four_clusters(memo):
    X = Id('X')
    Y = Id('Y')
    spn \
        = 0.25*(X >> norm(loc=0, scale=0.5) & Y >> norm(loc=0, scale=0.5)) \
        | 0.25*(X >> norm(loc=5, scale=0.5) & Y >> norm(loc=0, scale=0.5)) \
        | 0.25*(X >> norm(loc=0, scale=0.5) & Y >> norm(loc=5, scale=0.5)) \
        | 0.25*(X >> norm(loc=5, scale=0.5) & Y >> norm(loc=5, scale=0.5)) \

    A = X > 2
    B = Y > 2
    samples = spn.sample(100, prng)
    mi = spn.mutual_information(A, B, memo=memo)
    assert allclose(mi, 0)
    check_mi_properties(spn, A, B, memo)

    event = ((X>2) & (Y<2) | ((X<2) & (Y>2)))
    spn_condition = spn.condition(event)
    samples = spn_condition.sample(100, prng)
    assert all(event.evaluate(sample) for sample in samples)
    mi = spn_condition.mutual_information(X > 2, Y > 2)
    assert allclose(mi, log(2))

    check_mi_properties(spn, (X>1) | (Y<1), (Y>2), memo)
    check_mi_properties(spn, (X>1) | (Y<1), (X>1.5) & (Y>2), memo)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

from sppl.compilers.ast_to_spn import For
from sppl.compilers.ast_to_spn import Id
from sppl.compilers.ast_to_spn import IdArray
from sppl.compilers.ast_to_spn import IfElse
from sppl.compilers.ast_to_spn import Otherwise
from sppl.compilers.ast_to_spn import Sample
from sppl.compilers.ast_to_spn import Sequence
from sppl.distributions import bernoulli
from sppl.distributions import choice
from sppl.math_util import allclose

Y = Id('Y')
X = IdArray('X', 5)
Z = IdArray('Z', 5)

def test_simple_model():
    command = Sequence(
        Sample(Y, bernoulli(p=0.5)),
        For(0, 5, lambda i:
            Sample(X[i], bernoulli(p=1/(i+1)))))
    model = command.interpret()

    symbols = model.get_symbols()
    assert len(symbols) == 6
    assert Y in symbols
    assert X[0] in symbols
    assert X[1] in symbols
    assert X[2] in symbols
    assert X[3] in symbols
    assert X[4] in symbols
    assert model.logprob(X[0] << {1}) == log(1/1)
    assert model.logprob(X[1] << {1}) == log(1/2)
    assert model.logprob(X[2] << {1}) == log(1/3)
    assert model.logprob(X[3] << {1}) == log(1/4)
    assert model.logprob(X[4] << {1}) == log(1/5)

def test_complex_model():
    # Slow for larger number of repetitions
    # https://github.com/probcomp/sum-product-dsl/issues/43
    command = Sequence(
        Sample(Y, choice({'0': .2, '1': .2, '2': .2, '3': .2, '4': .2})),
        For(0, 3, lambda i: Sequence(
            Sample(Z[i], bernoulli(p=0.1)),
            IfElse(
                Y << {str(i)} | Z[i] << {0}, Sample(X[i], bernoulli(p=1/(i+1))),
                Otherwise,                   Sample(X[i], bernoulli(p=0.1))))))
    model = command.interpret()
    assert allclose(model.prob(Y << {'0'}), 0.2)

def test_complex_model_reorder():
    command = Sequence(
        Sample(Y, choice({'0': .2, '1': .2, '2': .2, '3': .2, '4': .2})),
        For(0, 3, lambda i:
            Sample(Z[i], bernoulli(p=0.1))),
        For(0, 3, lambda i:
            IfElse(
                Y << {str(i)}, Sample(X[i], bernoulli(p=1/(i+1))),
                Z[i] << {0},   Sample(X[i], bernoulli(p=1/(i+1))),
                Otherwise,     Sample(X[i], bernoulli(p=0.1)))))
    model = command.interpret()
    assert(allclose(model.prob(Y << {'0'}), 0.2))

def test_repeat_handcode_equivalence():
    model_repeat = make_model_for()
    model_hand = make_model_handcode()

    assert allclose(model_repeat.prob(Y << {'0', '1'}), 0.4)
    assert allclose(model_repeat.prob(Z[0] << {0}), 0.5)
    assert allclose(model_repeat.prob(Z[0] << {1}), 0.5)

    event_condition = (X[0] << {1}) | (Y << {'1'})
    model_repeat_condition = model_repeat.condition(event_condition)
    model_hand_condition = model_hand.condition(event_condition)

    for event in [
            Y << {'0','1'},
            Z[0] << {0},
            Z[1] << {0},
            X[0] << {0},
            X[1] << {0},
        ]:
        lp_repeat = model_repeat.logprob(event)
        lp_hand = model_hand.logprob(event)
        assert allclose(lp_hand, lp_repeat)

        lp_repeat_condition = model_repeat_condition.logprob(event)
        lp_hand_condition = model_hand_condition.logprob(event)
        assert allclose(lp_hand_condition, lp_repeat_condition)

    # This test case ensures that the duplicate subtrees in the mixture
    # components are pointers to the same object, and is obtained by
    # manually inspecting the rendering of the network generated by the
    # following code:
    #
    #   from sppl.magics.render import render_graphviz
    #   render_graphviz(model_repeat, show=True)
    #
    # See also test_cache_duplicate_subtrees.test_cache_complex_sum_of_product
    a = model_repeat.children[0].children[0].children[1].children[0]
    b = model_repeat.children[1].children[0]
    assert a is b

# ==============================================================================
# Helper functions.

def make_model_for(n=2):
    command = Sequence(
        Sample(Y, choice({'0': .2, '1': .2, '2': .2, '3': .2, '4': .2})),
        For(0, n, lambda i: Sequence(
            Sample(Z[i], bernoulli(p=.5)),
            IfElse(
                (Y << {str(i)}) | (Z[i] << {0}), Sample(X[i], bernoulli(p=.1)),
                Otherwise,                       Sample(X[i], bernoulli(p=.5))))))
    return command.interpret()

def make_model_handcode():
    command = Sequence(
        Sample(Y, choice({'0': .2, '1': .2, '2': .2, '3': .2, '4': .2})),
        Sample(Z[0], bernoulli(p=.5)),
        Sample(Z[1], bernoulli(p=.5)),
        IfElse(
            Y << {str(0)}, Sequence(
                Sample(X[0], bernoulli(p=.1)),
                IfElse(
                    Z[1] << {0},    Sample(X[1], bernoulli(p=.1)),
                    Otherwise,      Sample(X[1], bernoulli(p=.5)))),
            Y << {str(1)}, Sequence(
                Sample(X[1], bernoulli(p=.1)),
                IfElse(
                    Z[0] << {0},    Sample(X[0], bernoulli(p=.1)),
                    Otherwise,      Sample(X[0], bernoulli(p=.5)))),
            Otherwise, Sequence(
                IfElse(
                    Z[0] << {0},    Sample(X[0], bernoulli(p=.1)),
                    Otherwise,      Sample(X[0], bernoulli(p=.5))),
                IfElse(
                    Z[1] << {0},    Sample(X[1], bernoulli(p=.1)),
                    Otherwise,      Sample(X[1], bernoulli(p=.5))))))
    return command.interpret()



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

from sppl.distributions import gamma
from sppl.distributions import norm
from sppl.math_util import allclose
from sppl.spn import ProductSPN
from sppl.spn import SumSPN
from sppl.spn import spn_simplify_sum
from sppl.transforms import Id

def test_sum_simplify_nested_sum_1():
    X = Id('X')
    children = [
        SumSPN(
            [X >> norm(loc=0, scale=1), X >> norm(loc=0, scale=2)],
            [log(0.4), log(0.6)]),
        X >> gamma(loc=0, a=1),
    ]
    spn = SumSPN(children, [log(0.7), log(0.3)])
    assert spn.children == (
        children[0].children[0],
        children[0].children[1],
        children[1]
    )
    assert allclose(spn.weights[0], log(0.7) + log(0.4))
    assert allclose(spn.weights[1], log(0.7) + log(0.6))
    assert allclose(spn.weights[2], log(0.3))

def test_sum_simplify_nested_sum_2():
    X = Id('X')
    W = Id('W')
    children = [
        SumSPN([
            (X >> norm(loc=0, scale=1)) & (W >> norm(loc=0, scale=2)),
            (X >> norm(loc=0, scale=2)) & (W >> norm(loc=0, scale=1))],
            [log(0.9), log(0.1)]),
        (X >> norm(loc=0, scale=4)) & (W >> norm(loc=0, scale=10)),
        SumSPN([
            (X >> norm(loc=0, scale=1)) & (W >> norm(loc=0, scale=2)),
            (X >> norm(loc=0, scale=2)) & (W >> norm(loc=0, scale=1)),
            (X >> norm(loc=0, scale=8)) & (W >> norm(loc=0, scale=3)),],
            [log(0.4), log(0.3), log(0.3)]),
    ]
    spn = SumSPN(children, [log(0.4), log(0.4), log(0.2)])
    assert spn.children == (
        children[0].children[0],
        children[0].children[1],
        children[1],
        children[2].children[0],
        children[2].children[1],
        children[2].children[2],
    )
    assert allclose(spn.weights[0], log(0.4) + log(0.9))
    assert allclose(spn.weights[1], log(0.4) + log(0.1))
    assert allclose(spn.weights[2], log(0.4))
    assert allclose(spn.weights[3], log(0.2) + log(0.4))
    assert allclose(spn.weights[4], log(0.2) + log(0.3))
    assert allclose(spn.weights[5], log(0.2) + log(0.3))

def test_sum_simplify_leaf():
    Xd0 = Id('X') >> norm(loc=0, scale=1)
    Xd1 = Id('X') >> norm(loc=0, scale=2)
    Xd2 = Id('X') >> norm(loc=0, scale=3)
    spn = SumSPN([Xd0, Xd1, Xd2], [log(0.5), log(0.1), log(.4)])
    assert spn_simplify_sum(spn) == spn

    Xd0 = Id('X') >> norm(loc=0, scale=1)
    Xd1 = Id('X') >> norm(loc=0, scale=1)
    Xd2 = Id('X') >> norm(loc=0, scale=1)
    spn = SumSPN([Xd0, Xd1, Xd2], [log(0.5), log(0.1), log(.4)])
    assert spn_simplify_sum(spn) == Xd0

    Xd3 = Id('X') >> norm(loc=0, scale=2)
    spn = SumSPN([Xd0, Xd3, Xd1, Xd3], [log(0.5), log(0.1), log(.3), log(.1)])
    spn_simplified = spn_simplify_sum(spn)
    assert len(spn_simplified.children) == 2
    assert spn_simplified.children[0] == Xd0
    assert spn_simplified.children[1] == Xd3
    assert allclose(spn_simplified.weights[0], log(0.8))
    assert allclose(spn_simplified.weights[1], log(0.2))

def test_sum_simplify_product_collapse():
    A1 = Id('A') >> norm(loc=0, scale=1)
    A0 = Id('A') >> norm(loc=0, scale=1)
    B = Id('B') >> norm(loc=0, scale=1)
    B1 = Id('B') >> norm(loc=0, scale=1)
    B0 = Id('B') >> norm(loc=0, scale=1)
    C = Id('C') >> norm(loc=0, scale=1)
    C1 = Id('C') >> norm(loc=0, scale=1)
    D = Id('D') >> norm(loc=0, scale=1)
    spn = SumSPN([
        ProductSPN([A1, B, C, D]),
        ProductSPN([A0, B1, C, D]),
        ProductSPN([A0, B0, C1, D]),
    ], [log(0.4), log(0.4), log(0.2)])
    assert spn_simplify_sum(spn) == ProductSPN([A1, B, C, D])

def test_sum_simplify_product_complex():
    A1 = Id('A') >> norm(loc=0, scale=1)
    A0 = Id('A') >> norm(loc=0, scale=2)
    B = Id('B') >> norm(loc=0, scale=1)
    B1 = Id('B') >> norm(loc=0, scale=2)
    B0 = Id('B') >> norm(loc=0, scale=3)
    C = Id('C') >> norm(loc=0, scale=1)
    C1 = Id('C') >> norm(loc=0, scale=2)
    D = Id('D') >> norm(loc=0, scale=1)
    spn = SumSPN([
        ProductSPN([A1, B, C, D]),
        ProductSPN([A0, B1, C, D]),
        ProductSPN([A0, B0, C1, D]),
    ], [log(0.4), log(0.4), log(0.2)])

    spn_simplified = spn_simplify_sum(spn)
    assert isinstance(spn_simplified, ProductSPN)
    assert isinstance(spn_simplified.children[0], SumSPN)
    assert spn_simplified.children[1] == D

    ssc0 = spn_simplified.children[0]
    assert isinstance(ssc0.children[1], ProductSPN)
    assert ssc0.children[1].children == (A0, B0, C1)

    assert isinstance(ssc0.children[0], ProductSPN)
    assert ssc0.children[0].children[1] == C

    ssc0c0 = ssc0.children[0].children[0]
    assert isinstance(ssc0c0, SumSPN)
    assert isinstance(ssc0c0.children[0], ProductSPN)
    assert isinstance(ssc0c0.children[1], ProductSPN)
    assert ssc0c0.children[0].children == (A1, B)
    assert ssc0c0.children[1].children == (A0, B1)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

import pytest

import numpy

from sppl.distributions import choice
from sppl.distributions import gamma
from sppl.distributions import norm
from sppl.dnf import dnf_to_disjoint_union
from sppl.math_util import allclose
from sppl.math_util import isinf_neg
from sppl.math_util import logdiffexp
from sppl.math_util import lognorm
from sppl.math_util import logsumexp
from sppl.sets import Interval
from sppl.sets import inf as oo
from sppl.spn import LeafSPN
from sppl.spn import ProductSPN
from sppl.spn import SumSPN
from sppl.transforms import Exp
from sppl.transforms import Id
from sppl.transforms import Log

def test_product_distribution_normal_gamma_basic():
    X1 = Id('X1')
    X2 = Id('X2')
    X3 = Id('X3')
    X4 = Id('X4')
    children = [
        ProductSPN([
            X1 >> norm(loc=0, scale=1),
            X4 >> norm(loc=10, scale=1),
        ]),
        X2 >> gamma(loc=0, a=1),
        X3 >> norm(loc=2, scale=3)
    ]
    spn = ProductSPN(children)
    assert spn.children == (
        children[0].children[0],
        children[0].children[1],
        children[1],
        children[2],)
    assert spn.get_symbols() == frozenset([X1, X2, X3, X4])

    samples = spn.sample(2)
    assert len(samples) == 2
    for sample in samples:
        assert len(sample) == 4
        assert all([X in sample for X in (X1, X2, X3, X4)])

    samples = spn.sample_subset((X1, X2), 10)
    assert len(samples) == 10
    for sample in samples:
        assert len(sample) == 2
        assert X1 in sample
        assert X2 in sample

    samples = spn.sample_func(lambda X1, X2, X3: (X1, (X2**2, X3)), 1)
    assert len(samples) == 1
    assert len(samples[0]) == 2
    assert len(samples[0][1]) == 2

    with pytest.raises(ValueError):
        spn.sample_func(lambda X1, X5: X1 + X4, 1)

def test_product_inclusion_exclusion_basic():
    X = Id('X')
    Y = Id('Y')
    spn = ProductSPN([X >> norm(loc=0, scale=1), Y >> gamma(a=1)])

    a = spn.logprob(X > 0.1)
    b = spn.logprob(Y < 0.5)
    c = spn.logprob((X > 0.1) & (Y < 0.5))
    d = spn.logprob((X > 0.1) | (Y < 0.5))
    e = spn.logprob((X > 0.1) | ((Y < 0.5) & ~(X > 0.1)))
    f = spn.logprob(~(X > 0.1))
    g = spn.logprob((Y < 0.5) & ~(X > 0.1))

    assert allclose(a, spn.children[0].logprob(X > 0.1))
    assert allclose(b, spn.children[1].logprob(Y < 0.5))

    # Pr[A and B]  = Pr[A] * Pr[B]
    assert allclose(c, a + b)
     # Pr[A or B] = Pr[A] + Pr[B] - Pr[AB]
    assert allclose(d, logdiffexp(logsumexp([a, b]), c))
    # Pr[A or B] = Pr[A] + Pr[B & ~A]
    assert allclose(e, d)
    # Pr[A and B]  = Pr[A] * Pr[B]
    assert allclose(g, b + f)
    # Pr[A or (B & ~A)] = Pr[A] + Pr[B & ~A]
    assert allclose(e, logsumexp([a, b+f]))

    # (A => B) => Pr[A or B] = Pr[B]
    # i.e.,k (X > 1) => (X > 0).
    assert allclose(spn.logprob((X > 0) | (X > 1)), spn.logprob(X > 0))

    # Positive probability event.
    # Pr[A] = 1 - Pr[~A]
    event = ((0 < X) < 0.5) | ((Y < 0) & (1 < X))
    assert allclose(
        spn.logprob(event),
        logdiffexp(0, spn.logprob(~event)))

    # Probability zero event.
    event = ((0 < X) < 0.5) & ((Y < 0) | (1 < X))
    assert isinf_neg(spn.logprob(event))
    assert allclose(spn.logprob(~event), 0)

def test_product_condition_basic():
    X = Id('X')
    Y = Id('Y')
    spn = ProductSPN([X >> norm(loc=0, scale=1), Y >> gamma(a=1)])

    # Condition on (X > 0) and ((X > 0) | (Y < 0))
    # where the second clause reduces to first as Y < 0
    # has probability zero.
    for event in [(X > 0), (X  > 0) | (Y < 0)]:
        dX = spn.condition(event)
        assert isinstance(dX, ProductSPN)
        assert dX.children[0].symbol == Id('X')
        assert dX.children[0].conditioned
        assert dX.children[0].support == Interval.open(0, oo)
        assert dX.children[1].symbol == Id('Y')
        assert not dX.children[1].conditioned
        assert dX.children[1].Fl == 0
        assert dX.children[1].Fu == 1

    # Condition on (Y < 0.5)
    dY = spn.condition(Y < 0.5)
    assert isinstance(dY, ProductSPN)
    assert dY.children[0].symbol == Id('X')
    assert not dY.children[0].conditioned
    assert dY.children[1].symbol == Id('Y')
    assert dY.children[1].conditioned
    assert dY.children[1].support == Interval.Ropen(0, 0.5)

    # Condition on (X > 0) & (Y < 0.5)
    dXY_and = spn.condition((X > 0) & (Y < 0.5))
    assert isinstance(dXY_and, ProductSPN)
    assert dXY_and.children[0].symbol == Id('X')
    assert dXY_and.children[0].conditioned
    assert dXY_and.children[0].support == Interval.open(0, oo)
    assert dXY_and.children[1].symbol == Id('Y')
    assert dXY_and.children[1].conditioned
    assert dXY_and.children[1].support == Interval.Ropen(0, 0.5)

    # Condition on (X > 0) | (Y < 0.5)
    event = (X > 0) | (Y < 0.5)
    dXY_or = spn.condition((X > 0) | (Y < 0.5))
    assert isinstance(dXY_or, SumSPN)
    assert all(isinstance(d, ProductSPN) for d in dXY_or.children)
    assert allclose(dXY_or.logprob(X > 0),dXY_or.weights[0])
    samples = dXY_or.sample(100, prng=numpy.random.RandomState(1))
    assert all(event.evaluate(sample) for sample in samples)

    # Condition on a disjoint union with one term in second clause.
    dXY_disjoint_one = spn.condition((X > 0) & (Y < 0.5) | (X <= 0))
    assert isinstance(dXY_disjoint_one, SumSPN)
    component_0 = dXY_disjoint_one.children[0]
    assert component_0.children[0].symbol == Id('X')
    assert component_0.children[0].conditioned
    assert component_0.children[0].support == Interval.open(0, oo)
    assert component_0.children[1].symbol == Id('Y')
    assert component_0.children[1].conditioned
    assert component_0.children[1].support == Interval.Ropen(0, 0.5)
    component_1 = dXY_disjoint_one.children[1]
    assert component_1.children[0].symbol == Id('X')
    assert component_1.children[0].conditioned
    assert component_1.children[0].support == Interval(-oo, 0)
    assert component_1.children[1].symbol == Id('Y')
    assert not component_1.children[1].conditioned

    # Condition on a disjoint union with two terms in each clause
    dXY_disjoint_two = spn.condition((X > 0) & (Y < 0.5) | ((X <= 0) & ~(Y < 3)))
    assert isinstance(dXY_disjoint_two, SumSPN)
    component_0 = dXY_disjoint_two.children[0]
    assert component_0.children[0].symbol == Id('X')
    assert component_0.children[0].conditioned
    assert component_0.children[0].support == Interval.open(0, oo)
    assert component_0.children[1].symbol == Id('Y')
    assert component_0.children[1].conditioned
    assert component_0.children[1].support == Interval.Ropen(0, 0.5)
    component_1 = dXY_disjoint_two.children[1]
    assert component_1.children[0].symbol == Id('X')
    assert component_1.children[0].conditioned
    assert component_1.children[0].support == Interval(-oo, 0)
    assert component_1.children[1].symbol == Id('Y')
    assert component_1.children[1].conditioned
    assert component_1.children[1].support == Interval(3, oo)

    # Some various conditioning.
    spn.condition((X > 0) & (Y < 0.5) | ((X <= 1) | ~(Y < 3)))
    spn.condition((X > 0) & (Y < 0.5) | ((X <= 1) & (Y < 3)))

def test_product_condition_or_probabilithy_zero():
    X = Id('X')
    Y = Id('Y')
    spn = ProductSPN([X >> norm(loc=0, scale=1), Y >> gamma(a=1)])

    # Condition on event which has probability zero.
    event = (X > 2) & (X < 2)
    with pytest.raises(ValueError):
        spn.condition(event)
    assert spn.logprob(event) == -float('inf')

    # Condition on event which has probability zero.
    event = (Y < 0) | (Y < -1)
    with pytest.raises(ValueError):
        spn.condition(event)
    assert spn.logprob(event) == -float('inf')
    # Condition on an event where one clause has probability
    # zero, yielding a single product.
    spn_condition = spn.condition((Y < 0) | ((Log(X) >= 0) & (1 <= Y)))
    assert isinstance(spn_condition, ProductSPN)
    assert spn_condition.children[0].symbol == X
    assert spn_condition.children[0].conditioned
    assert spn_condition.children[0].support == Interval(1, oo)
    assert spn_condition.children[1].symbol == Y
    assert spn_condition.children[1].conditioned
    assert spn_condition.children[0].support == Interval(1, oo)

    # We have (X < 2) & ~(1 < exp(|3X**2|) is empty.
    # Thus Y remains unconditioned,
    #   and X is partitioned into (-oo, 0) U (0, oo) with equal weight.
    event = (Exp(abs(3*X**2)) > 1) | ((Log(Y) < 0.5) & (X < 2))
    spn_condition = spn.condition(event)
    #
    # The most concise representation of spn_condition is:
    #   (Product (Sum [.5 .5] X|X<0 X|X>0) Y)
    assert isinstance(spn_condition, ProductSPN)
    assert isinstance(spn_condition.children[0], SumSPN)
    assert spn_condition.children[0].weights == (-log(2), -log(2))
    assert spn_condition.children[0].children[0].conditioned
    assert spn_condition.children[0].children[1].conditioned
    assert spn_condition.children[0].children[0].support \
        in [Interval.Ropen(-oo, 0), Interval.Lopen(0, oo)]
    assert spn_condition.children[0].children[1].support \
        in [Interval.Ropen(-oo, 0), Interval.Lopen(0, oo)]
    assert spn_condition.children[0].children[0].support \
        != spn_condition.children[0].children[1].support
    assert spn_condition.children[1].symbol == Y
    assert not spn_condition.children[1].conditioned
    #
    # However, now we solve the constraint and factor the query into
    # a disjoint union at the root of the SPN, not once per node, we
    # no longer have guarantees of generating the smallest network.
    # In this case, the answer is the more verbose:
    #   (Sum [.5 .5] (Product X|X<0 Y) (Product X|X>0 Y)
    # assert isinstance(spn_condition, SumSPN)
    # assert spn_condition.weights == (-log(2), -log(2))
    # assert isinstance(spn_condition.children[0], ProductSPN)
    # assert X \
    #     == spn_condition.children[0].children[0].symbol \
    #     == spn_condition.children[1].children[0].symbol
    # assert spn_condition.children[0].children[0].conditioned
    # assert spn_condition.children[1].children[0].conditioned
    # assert spn_condition.children[0].children[0].support \
    #     == Interval.Ropen(-oo, 0)
    # assert spn_condition.children[1].children[0].support \
    #     == Interval.Lopen(0, oo)
    # assert Y \
    #     == spn_condition.children[0].children[1].symbol \
    #     == spn_condition.children[1].children[1].symbol
    # assert not spn_condition.children[0].children[1].conditioned
    # assert not spn_condition.children[1].children[1].conditioned

def test_product_disjoint_union_numerical():
    X = Id('X')
    Y = Id('Y')
    Z = Id('Z')
    spn = ProductSPN([
        X >> norm(loc=0, scale=1),
        Y >> norm(loc=0, scale=2),
        Z >> norm(loc=0, scale=2),
    ])

    for event in [
        (1/X < 4) | (X > 7),
        (2*X-3 > 0) | (Log(Y) < 3),
        ((X > 0) & (Y < 1)) | ((X < 1) & (Y < 3)) | ~(X<<{1, 2}),
        ((X > 0) & (Y < 1)) | ((X < 1) & (Y < 3)) | (Z < 0),
        ((X > 0) & (Y < 1)) | ((X < 1) & (Y < 3)) | (Z < 0) | ~(X <<{1, 3}),
    ]:
        clauses = dnf_to_disjoint_union(event)
        logps = [spn.logprob(s) for s in clauses.subexprs]
        assert allclose(logsumexp(logps), spn.logprob(event))

def test_product_disjoint_union_nominal():
    N = Id('N')
    P = Id('P')

    nationality = N >> choice({'India': 0.5, 'USA': 0.5})
    perfect = P >> choice({'Imperfect': 0.99, 'Perfect': 0.01})
    student = nationality & perfect

    condition_1 = (N << {'India'}) & (P << {'Imperfect'})
    condition_2 = (N << {'India'}) & (P << {'Perfect'})
    condition_3 = (N << {'USA'}) & (P << {'Imperfect'})
    condition_4 = (N << {'USA'}) & (P << {'Perfect'})

    event_1 = condition_1
    event_2 = condition_2 & ~condition_1
    event_3 = condition_3 & ~condition_2 & ~condition_1
    event_4 = condition_4 & ~condition_3 & ~condition_2 & ~condition_1

    assert allclose(student.prob(event_1), 0.5*0.99)
    assert allclose(student.prob(event_2), 0.5*0.01)
    assert allclose(student.prob(event_3), 0.5*0.99)
    assert allclose(student.prob(event_4), 0.5*0.01)

A = Id('A')
B = Id('B')
C = Id('C')
D = Id('D')
spn_abcd \
    = norm(loc=0, scale=1)(A) \
    & norm(loc=0, scale=1)(B) \
    & norm(loc=0, scale=1)(C) \
    & norm(loc=0, scale=1)(D)
def test_product_condition_simplify_a():
    spn = spn_abcd.condition((A > 1) | (A < -1))
    assert isinstance(spn, ProductSPN)
    assert spn_abcd.children[1] in spn.children
    assert spn_abcd.children[2] in spn.children
    assert spn_abcd.children[3] in spn.children
    idx_sum = [i for i, c in enumerate(spn.children) if isinstance(c, SumSPN)]
    assert len(idx_sum) == 1
    assert allclose(spn.children[idx_sum[0]].weights[0], -log(2))
    assert allclose(spn.children[idx_sum[0]].weights[1], -log(2))
    assert spn.children[idx_sum[0]].children[0].conditioned
    assert spn.children[idx_sum[0]].children[1].conditioned

def test_product_condition_simplify_ab():
    spn = spn_abcd.condition((A > 1) | (B < 0))
    assert isinstance(spn, ProductSPN)
    assert spn_abcd.children[2] in spn.children
    assert spn_abcd.children[2] in spn.children
    idx_sum = [i for i, c in enumerate(spn.children) if isinstance(c, SumSPN)]
    assert len(idx_sum) == 1
    spn_sum = spn.children[idx_sum[0]]
    assert isinstance(spn_sum.children[0], ProductSPN)
    assert isinstance(spn_sum.children[1], ProductSPN)
    lp0 = spn_abcd.logprob(A > 1)
    lp1 = spn_abcd.logprob((B < 0) & ~(A > 1))
    weights = lognorm([lp0, lp1])
    assert allclose(spn_sum.weights[0], weights[0])
    assert allclose(spn_sum.weights[1], weights[1])

def test_product_condition_simplify_abc():
    spn = spn_abcd.condition((A > 1) | (B < 0) | (C > 2))
    assert isinstance(spn, ProductSPN)
    assert len(spn.children) == 2
    assert spn_abcd.children[3] in spn.children
    idx_sum = [i for i, c in enumerate(spn.children) if isinstance(c, SumSPN)]
    assert len(idx_sum) == 1
    spn_sum = spn.children[idx_sum[0]]
    assert len(spn_sum.children) == 2
    lp0 = spn_abcd.logprob(A > 1)
    lp1 = spn_abcd.logprob((B < 0) & ~(A > 1))
    lp2 = spn_abcd.logprob((C > 2) & ~(B < 0) & ~(A > 1))
    weights = lognorm([lp0, lp1, lp2])
    assert allclose(spn_sum.weights[1], weights[-1])
    assert isinstance(spn_sum.children[0], ProductSPN)
    assert isinstance(spn_sum.children[0].children[0], SumSPN)
    assert isinstance(spn_sum.children[0].children[1], LeafSPN)
    assert isinstance(spn_sum.children[1], ProductSPN)
    assert isinstance(spn_sum.children[1].children[0], LeafSPN)
    assert isinstance(spn_sum.children[1].children[1], LeafSPN)
    assert isinstance(spn_sum.children[1].children[2], LeafSPN)
